# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import numpy as np
import logging
import shutil
from osgeo.gdalconst import GA_ReadOnly
from osgeo import gdal
from equi7grid.equi7grid import Equi7Grid
from equi7grid.image2equi7grid import image2equi7grid
from copy import deepcopy
from arepytools.timing.precisedatetime import PreciseDateTime

# biomassL2 processor imports
from biopal.fd.processing_FD import alg_wishart_SU
from biopal.data_operations.data_operations import (
    read_and_oversample_data,
    read_and_oversample_aux_data,
    fnf_equi7_load_filter_equi7format,
    fnf_tandemx_load_filter_equi7format,
    apply_dem_flattening,
    get_equi7_tiff_names,
)
from biopal.utility.utility_functions import (
    start_logging,
    Task,
    choose_equi7_sampling,
    check_if_path_exists,
    check_if_geometry_auxiliaries_are_present,
    check_fnf_folder_format,
    get_min_time_stamp_repository,
    resolution_heading_correction,
    decode_unique_acquisition_id_string,
    save_breakpoints,
    set_gdal_paths,
    get_data_time_stamp,
    check_equi7_mask_coverage,
)
from biopal.statistics.utility_statistics import (
    main_covariance_estimation_SR,
    covariance_matrix_mat2vec,
    covariance_matrix_vec2mat,
)

from biopal.geocoding.geocoding import (
    geocoding,
    geocoding_init,
)
from biopal.io.xml_io import (
    parse_input_file,
    parse_configuration_file,
    write_input_file,
    core_processing_fd,
)

from biopal.io.data_io import tiff_formatter
from biopal.screen_calibration.screen_calibration import apply_calibration_screens
from biopal.geometry.utility_geometry import compute_and_oversample_geometry_auxiliaries
from biopal.ground_cancellation.ground_cancellation import ground_cancellation


class InvalidInputError(ValueError):
    pass


class ForestDisturbance(Task):
    """FD main APP ForestDisturbance
    
    run this APP to execute the complete Forest Disturbance processing chain. 

    ForestDisturbance is composed by two sub APPS automatically called in sequence  
    when standard launch is performed:
    InitializeAlgorithmFD -> CoreProcessingFD

    Refer to dataset_query, InitializeAlgorithmFD and CoreProcessingFD documentation for step by step run.
    
    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file

    Methods
    -------
    run( input_file_path )
        run the ForestDisturbance processing
    name : str
        name of the APP
        
    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query: it's the APP to be called before this APP
    InitializeAlgorithmFD : it's the first of the two sub-APPs called by ForestDisturbance preocessor
    CoreProcessingFD : it's the second of the two sub-APPs called by ForestDisturbance processor

    Examples
    --------
    Manual FD chain execution

    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.fd.main_FD import ForestDisturbance
    >>> dq_obj = dataset_query()
    >>> input_file_up = dq_obj.run( input_file )
    >>> chain_obj = ForestDisturbance( configuration_file )
    >>> chain_obj.run( input_file_up )

    - input_file: path of the BioPAL input file
    - input_file_up: same of input_file with also the "stack_based_processing" section
    - configuration_file: path of the BioPAL configuration file
    """

    def __init__(
        self, configuration_file,
    ):
        super().__init__(configuration_file)

    def _run(self, input_file):

        # Main APP #1: Prepare initialization APP
        #              organize the input dataSet by grouping all the temporal global cycles of each stack
        initialize_fd_obj = InitializeAlgorithmFD(self.configuration_file)

        # Run Main APP #1: the FD initialization APP
        input_file_updated = initialize_fd_obj.run(input_file)

        # Main APP #2: FD Core Processing (change detection for each nominal geometry)
        fd_processing_obj = CoreProcessingFD(self.configuration_file)

        # Run Main APP #2: FD Core Processing
        fd_processing_obj.run(input_file_updated)


class InitializeAlgorithmFD(Task):
    """InitializeAlgorithmFD APP

    InitializeAlgorithmFD APP is the first of the two sub-APPs called by ForestDisturbance processor.

    It organizes the input dataSet by grouping all the different temporal global 
    cycles of each stack with same nominal geometry, preparing the inputs for 
    the "CoreProcessingFD" APP.
    It is automatically called in the default run (see Getting Started section), 
    or can be manually launched (stand alone) following this sequence: 
    dataset_query -> InitializeAlgorithmFD -> CoreProcessingFD

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file

    Methods
    -------
    run( input_file_path )
        run the InitializeAlgorithmFD process
    name : str
        name of the APP

    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query: it's the APP to be called before this APP
    CoreProcessingFD : it's the second of the two sub-APPs called by ForestDisturbance processor

    Examples
    --------
    Manual FD chain execution
        
    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.fd.main_FD import InitializeAlgorithmFD, CoreProcessingFD
    >>> dq_obj = dataset_query()
    >>> input_file_up = dq_obj.run( input_file )
    >>> init_fd_obj = InitializeAlgorithmFD( config_file )
    >>> input_file_up2 = init_fd_obj.run( input_file_up )
    >>> fdp_obj = CoreProcessingFD( config_file )
    >>> fdp_obj.run( input_file_up2 )
    
    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_fd" section
    - config_file: path of the BioPAL configuration file
    """

    def __init__(self, configuration_file):

        super().__init__(configuration_file)

    def _run(self, input_file):

        ########################## INITIAL STEPS ##############################

        # parse the input file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(input_file, "FILE")
            input_params_obj = parse_input_file(input_file)
        else:
            input_params_obj = input_file
        if (
            input_params_obj.L2_product is None
            or input_params_obj.output_specification is None
            or input_params_obj.dataset_query is None
            or input_params_obj.stack_based_processing is None
        ):
            error_message = [
                "InitializeAlgorithmFD APP input file should contain at least"
                ' "L2_product", "output_specification", "dataset_query" and "stack_based_processing" sections'
            ]
            raise InvalidInputError(error_message)

        # parse the configuration file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(self.configuration_file, "FILE")
            conf_params_obj = parse_configuration_file(self.configuration_file)
        else:
            conf_params_obj = self.configuration_file
        if conf_params_obj.processing_flags is None or conf_params_obj.change_detection_fd is None:
            error_message = [
                "Configuration file for  InitializeAlgorithmFD APP should contain at least"
                ' "processing_flags" and "change_detection_fd" sections'
            ]
            raise InvalidInputError(error_message)

        # Initialize the logger, of the APP has been called in a stand-alone call
        if not hasattr(logging.getLoggerClass().root.handlers[0], "baseFilename"):
            start_logging(
                os.path.dirname(input_params_obj.output_specification.output_folder),
                input_params_obj.L2_product,
                "DEBUG",
                app_name="InitializeAlgorithmFD",
            )
        logging.info("FD initialize algorithm APP starting\n")

        ### managing output folders:
        products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
        if conf_params_obj.processing_flags.save_breakpoints:
            breakpoints_output_folder = os.path.join(products_folder, "breakpoints")
            logging.info("FD: Breakpoints will be saved into: " + breakpoints_output_folder)
            os.makedirs(breakpoints_output_folder)

        temp_output_folder = os.path.join(products_folder, "temp")
        logging.info("FD: Temporary data folder:" + temp_output_folder + "\n")

        os.makedirs(temp_output_folder)

        ### grouping temoral global cycles for each stack
        cycles_composition = (
            {}
        )  # dict which groupes togheter all the differente global cycles for stacks with same nominal geometry
        # cycles_composition[nominal_geometry_stack_id][global_cycle_number]
        logging.info('FD: grouping togheter the stacks which have same "nominal geometry" in different global cycles:')
        for idx, (unique_stack_id, unique_acq_pf_names) in enumerate(
            input_params_obj.stack_based_processing.stack_composition.items()
        ):
            nominal_geometry_stack_id = unique_stack_id[
                6:
            ]  # remove the global cycle indication, maintain the geometry indications
            global_cycle_idx = int(unique_stack_id[3:5])  # \ the cycle number of curent uniqie stack

            if not nominal_geometry_stack_id in cycles_composition.keys():
                cycles_composition[nominal_geometry_stack_id] = {}

            cycles_composition[nominal_geometry_stack_id][global_cycle_idx] = unique_acq_pf_names
        logging.info(
            "    for the #{} stacks in input, #{} different nominal geometries have been found \n".format(
                len(input_params_obj.stack_based_processing.stack_composition), len(cycles_composition)
            )
        )

        # write the input file with the sections needed by the Core Processing FD APP:
        out_input_file_xml = os.path.join(
            os.path.dirname(input_params_obj.output_specification.output_folder), "Input_File_CoreProcessingFD.xml"
        )
        input_params_obj.core_processing_fd = fill_core_processing_fd_obj(cycles_composition)
        write_input_file(input_params_obj, out_input_file_xml)

        ########################## INITIAL STEPS END #############################

        return out_input_file_xml


class CoreProcessingFD(Task):
    """CoreProcessingFD APP

    CoreProcessingFD is the second of the two sub-APPs called by ForestDisturbance processor.

    It calls the "change detection" algorithm (which is implemented as a function) 
    once for each nominal geometry:
    this app will cycle over each nominal geometry and for each of them it calls the change detection 
    function that will cycle over all the global cycles, computing the change detection for the current geometry.
    It is automatically called in the default run (see Getting Started section), 
    or can be manually launched (stand alone) following this sequence: 
    dataset_query -> InitializeAlgorithmFD -> CoreProcessingFD

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file

    Methods
    -------
    run( input_file_path )
        run the InitializeAlgorithmFD process
    name : str
        name of the APP
                
    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query: it’s the first APP to be called in the manual sequence
    InitializeAlgorithmFD : it’s the APP which prepares the input for this APP

    Notes
    -----
    It is possible to launch many istances of "CoreProcessingFD" APP without need to re launch all the processor.
    
    Examples
    --------
    Manual FD chain execution
        
    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.fd.main_FD import InitializeAlgorithmFD, CoreProcessingFD
    >>> dq_obj = dataset_query()
    >>> input_file_up = dq_obj.run( input_file )
    >>> init_fd_obj = InitializeAlgorithmFD( config_file )
    >>> input_file_up2 = init_fd_obj.run( input_file_up )
    >>> fdp_obj = CoreProcessingFD( config_file )
    >>> fdp_obj.run( input_file_up2 )
    
    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_fd" section
    - config_file: path of the BioPAL configuration file          
    """

    def __init__(self, configuration_file):

        super().__init__(configuration_file)

    def _run(self, input_file):

        # parse the input file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(input_file, "FILE")
            input_params_obj = parse_input_file(input_file)
        else:
            input_params_obj = input_file
        if (
            input_params_obj.L2_product is None
            or input_params_obj.output_specification is None
            or input_params_obj.dataset_query is None
            or input_params_obj.core_processing_fd is None
        ):
            error_message = [
                "CoreProcessingFD APP input file should contain at least"
                ' "L2_product", "output_specification", "dataset_query" and "core_processing_fd" sections'
            ]
            raise InvalidInputError(error_message)

        # parse the configuration file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(self.configuration_file, "FILE")
            conf_params_obj = parse_configuration_file(self.configuration_file)
        else:
            conf_params_obj = self.configuration_file
        if (
            conf_params_obj.processing_flags is None
            or conf_params_obj.ground_cancellation is None
            or conf_params_obj.change_detection_fd is None
        ):
            error_message = [
                "Configuration file for CoreProcessingFD APP should contain at least"
                ' "processing_flags", "ground_cancellation" and "change_detection_fd" sections'
            ]
            raise InvalidInputError(error_message)

        # Initialize the logger, of the APP has been called in a stand-alone call
        if not hasattr(logging.getLoggerClass().root.handlers[0], "baseFilename"):
            start_logging(
                os.path.dirname(input_params_obj.output_specification.output_folder),
                input_params_obj.L2_product,
                "DEBUG",
                app_name="CoreProcessingFD",
            )
        logging.info("FD core-processing APP starting\n")

        ### initialize the equi7 sampling grid
        equi7_sampling = choose_equi7_sampling(
            conf_params_obj.change_detection_fd.product_resolution,
            input_params_obj.output_specification.geographic_grid_sampling,
        )
        logging.info("EQUI7 Grid sampling used: {}".format(equi7_sampling))

        ### get temporal date time of the input data (get the minimum date from all the stacks)
        time_tag_mjd_initial = get_min_time_stamp_repository(
            input_params_obj.dataset_query.L1C_repository, input_params_obj.stack_based_processing.stack_composition
        )
        # cycle over each nominal geometry stack (containing all its global cycles)
        final_forest_mask_data = {}  # one mask for each equi7 tile, it is updated at each stack-cycle
        for (
            nominal_geometry_stack_id,
            global_cycle_dict,
        ) in input_params_obj.core_processing_fd.cycles_composition.items():

            (final_forest_mask_data, temp_output_folder,) = changeDetection(  # this goes in input to next iteration
                self.configuration_file,
                final_forest_mask_data,
                nominal_geometry_stack_id,
                global_cycle_dict,
                time_tag_mjd_initial,
                equi7_sampling,
                input_params_obj,
                conf_params_obj,
            )

        # Ending the ForestDisturbance APP
        if conf_params_obj.processing_flags.delete_temporary_files:
            try:
                shutil.rmtree(temp_output_folder)
            except:
                pass
        logging.info("FD processor ended correctly.\n")


def changeDetection(
    configuration_file,
    final_forest_mask_data,
    nominal_geometry_stack_id,
    global_cycle_dict,
    time_tag_mjd_initial,
    equi7_sampling,
    input_params_obj,
    conf_params_obj,
):

    """
    This is the core of the FD algorithm
    """

    # get needed parameters from input and configuration files
    geographic_boundaries = input_params_obj.stack_based_processing.geographic_boundaries

    gdal_path, _ = set_gdal_paths(conf_params_obj.gdal.gdal_path, conf_params_obj.gdal.gdal_environment_path)

    ########################## INITIALIZATIONS ############################
    number_of_pols = 3

    products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
    temp_output_folder = os.path.join(products_folder, "temp")
    if not os.path.exists(temp_output_folder):
        os.makedirs(temp_output_folder)
    if conf_params_obj.processing_flags.save_breakpoints:
        breakpoints_output_folder = os.path.join(products_folder, "breakpoints")
        if not os.path.exists(breakpoints_output_folder):
            os.makedirs(breakpoints_output_folder)

    ### initialize the equi7 sampling grid
    e7g = Equi7Grid(equi7_sampling)

    _, heading_deg, rg_swath_idx, rg_sub_swath_idx, az_swath_idx, _ = decode_unique_acquisition_id_string(
        next(iter(global_cycle_dict.values()))[0]
    )
    logging.info("FD: computing disturbance for following nominal geometry " + nominal_geometry_stack_id + ":")
    logging.info(
        "Heading = {}, RG swath = {} RG sub-swath = {}, AZ swath = {}".format(
            heading_deg, rg_swath_idx, rg_sub_swath_idx, az_swath_idx
        )
    )

    number_of_global_cycles = len(global_cycle_dict.keys())
    logging.info("FD: #{} global cycles found for the current nominal geometry".format(number_of_global_cycles))
    ######################### INITIALIZATIONS END #########################

    # cycle over all global cycles of current nominal geometry (current stack)
    for time_step_idx, (global_cycle_idx, uniqie_acq_ids_all_cycles_list) in enumerate(global_cycle_dict.items()):

        time_tag_mjd_curr = get_data_time_stamp(
            input_params_obj.dataset_query.L1C_repository, uniqie_acq_ids_all_cycles_list[0]
        )

        # reconstruct current unique_stack_id
        global_cycle_idx_str = str(global_cycle_idx)
        if len(global_cycle_idx_str) == 1:
            global_cycle_idx_str = "0" + global_cycle_idx_str
        unique_stack_id = "GC_" + global_cycle_idx_str + "_" + nominal_geometry_stack_id

        if time_step_idx == 0:
            global_cycle_idx_prev_str = str(global_cycle_idx - 1)
        else:
            global_cycle_idx_prev_str = str([*global_cycle_dict.keys()][time_step_idx - 1])
            logging.debug(
                "time_step_idx {}, global_cycle_idx_prev_str {} ".format(time_step_idx, global_cycle_idx_prev_str)
            )

        if len(global_cycle_idx_prev_str) == 1:
            global_cycle_idx_prev_str = "0" + global_cycle_idx_prev_str

        unique_stack_id_prev = "GC_" + global_cycle_idx_prev_str + "_" + nominal_geometry_stack_id

        logging.info(
            "FD: disturbance computation step #{} of #{}; current global cycle is ".format(
                time_step_idx + 1, number_of_global_cycles
            )
            + '"GC_'
            + global_cycle_idx_str
            + '"'
            + " current geometry is "
            + nominal_geometry_stack_id
        )

        # make temporary sub-directories
        temp_output_folder_sr = os.path.join(temp_output_folder, "slantRange", unique_stack_id)
        temp_output_folder_gr = os.path.join(temp_output_folder, "geocoded", unique_stack_id)
        temp_output_folder_e7 = os.path.join(temp_output_folder, "equi7", unique_stack_id)
        os.makedirs(temp_output_folder_sr)
        os.makedirs(temp_output_folder_gr)
        os.makedirs(temp_output_folder_e7)

        ### load data ( and oversample if requested and if needed )
        try:
            logging.info("FD: input data reading:...")

            beta0_calibrated, master_id, raster_info, raster_info_orig = read_and_oversample_data(
                input_params_obj.dataset_query.L1C_repository,
                uniqie_acq_ids_all_cycles_list,
                conf_params_obj.processing_flags.enable_resampling,
            )

        except Exception as e:
            logging.error("FD: error during input data reading: " + str(e), exc_info=True)
            raise
        logging.info("...done.\n")

        ### load or compute auxiliary data
        try:

            read_ref_h = (
                not conf_params_obj.processing_flags.apply_calibration_screen
                and conf_params_obj.processing_flags.DEM_flattening
            )
            read_cal_screens = conf_params_obj.processing_flags.apply_calibration_screen
            geometry_aux_are_present = check_if_geometry_auxiliaries_are_present(
                input_params_obj.stack_based_processing,
                unique_stack_id,
                uniqie_acq_ids_all_cycles_list,
                read_ref_h=read_ref_h,
            )

            if conf_params_obj.processing_flags.compute_geometry or not geometry_aux_are_present:

                # messages for the log:
                if conf_params_obj.processing_flags.compute_geometry:
                    logging.info("FD: calling geometry library for stack " + unique_stack_id + "\n")
                    if geometry_aux_are_present:
                        logging.warning("    geometry auxiliaries will be overwritten for stack " + unique_stack_id)
                    else:
                        logging.info("\n")
                else:
                    logging.warning(
                        'FD: calling geometry library since AuxiliaryProductsFolder "Geometry" is empty or not complete \n'
                    )

                _, _, ellipsoid_slope, _, _, _, sar_geometry_master = compute_and_oversample_geometry_auxiliaries(
                    input_params_obj.dataset_query.L1C_repository,
                    input_params_obj.stack_based_processing,
                    unique_stack_id,
                    uniqie_acq_ids_all_cycles_list,
                    master_id,
                    conf_params_obj.processing_flags.enable_resampling,
                    force_ellipsoid=True,
                )

                (
                    ecef_grid,
                    off_nadir_angle_rad,
                    slope,
                    kz,
                    reference_height,
                    _,
                    _,
                ) = compute_and_oversample_geometry_auxiliaries(
                    input_params_obj.dataset_query.L1C_repository,
                    input_params_obj.stack_based_processing,
                    unique_stack_id,
                    uniqie_acq_ids_all_cycles_list,
                    master_id,
                    conf_params_obj.processing_flags.enable_resampling,
                    sar_geometry_master=sar_geometry_master,
                )

                logging.info("Geometry library: correcting geometry global / local reference")
                slope = slope - ellipsoid_slope
                for swath_id in kz.keys():
                    kz[swath_id] = (
                        kz[swath_id]
                        * np.sin(off_nadir_angle_rad[master_id])
                        / np.sin(off_nadir_angle_rad[master_id] - ellipsoid_slope)
                    )
                off_nadir_angle_rad[master_id] = off_nadir_angle_rad[master_id] - ellipsoid_slope
                del ellipsoid_slope

                logging.info("FD: ...geometry auxiliaries computation done.")

            else:

                logging.info(
                    "FD: geometry auxiliaries are provided from user, so they are now being loaded and not computed, for stack "
                    + unique_stack_id
                    + "\n"
                )

                (
                    ecef_grid,
                    off_nadir_angle_rad,
                    slope,
                    kz,
                    reference_height,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                ) = read_and_oversample_aux_data(
                    input_params_obj.stack_based_processing,
                    unique_stack_id,
                    uniqie_acq_ids_all_cycles_list,
                    conf_params_obj.processing_flags.enable_resampling,
                    raster_info_orig,
                )
                logging.info("FD: ...geometry auxiliaries loading done.")

            # read the rest of auxiliaries which are notpart of the geometry library:
            if read_cal_screens:

                logging.warning("FD: loading calibration screens \n")

                _, _, _, _, _, _, _, cal_screens, cal_screens_raster_info, _, _, _ = read_and_oversample_aux_data(
                    input_params_obj.stack_based_processing,
                    unique_stack_id,
                    uniqie_acq_ids_all_cycles_list,
                    conf_params_obj.processing_flags.enable_resampling,
                    raster_info_orig,
                    read_cal_screens=read_cal_screens,
                    read_ecef=False,
                    read_off_nadir=False,
                    read_slope=False,
                    read_kz=False,
                    read_ref_h=False,
                    read_dist=False,
                )
                logging.info("...done.\n")

        except Exception as e:
            logging.error("FD: error during auxiliary data computation and/or loading: " + str(e), exc_info=True)
            raise

        ### Screen calibration (ground steering)
        try:
            if conf_params_obj.processing_flags.apply_calibration_screen:
                logging.info("FH: applying calibration screen...")
                beta0_calibrated = apply_calibration_screens(
                    beta0_calibrated, raster_info, cal_screens, cal_screens_raster_info, master_id
                )
                logging.info("...done.\n")

            elif conf_params_obj.processing_flags.DEM_flattening:
                logging.info("FH: DEM flattening... ")
                beta0_calibrated = apply_dem_flattening(beta0_calibrated, kz, reference_height, master_id, raster_info)
                logging.info("...done.\n")

        except Exception as e:
            logging.error("FH: error during screen calibration or DEM flattening." + str(e), exc_info=True)
            raise

        ### compute mean incidence angle
        look_angle_rad = np.nanmean(off_nadir_angle_rad[master_id])  # 0.4886921905584123
        logging.info("FD: incidence angle used is {} [deg] \n".format(np.rad2deg(look_angle_rad)))

        ### ground notching
        try:
            logging.info("FD: ground contribute cancellation...:")

            beta0_notched = ground_cancellation(
                beta0_calibrated,
                kz,
                conf_params_obj.ground_cancellation.multi_master_flag,
                conf_params_obj.ground_cancellation.enhanced_forest_height,
                conf_params_obj.ground_cancellation.equalization_flag,
                raster_info.resolution_m_slant_rg,
                off_nadir_angle_rad[master_id],
                slope,
            )  # beta0_calibrated: 3 pol (nrg x Naz  x Nimmagini); beta0_notched: 3 pol (Nrg x Naz x 1 immagine)

            del beta0_calibrated, kz

        except Exception as e:
            logging.error("FD: error during ground cancellation: " + str(e), exc_info=True)
            raise
        logging.info("...done.\n")

        ### saving breakpoints
        if conf_params_obj.processing_flags.save_breakpoints:
            logging.info("FD: saving breakpoints (in slant range geometry) on " + breakpoints_output_folder)
            post_string = "_SR_" + unique_stack_id

            breakpoint_names = ["ground_cancelled" + post_string]

            save_breakpoints(breakpoints_output_folder, breakpoint_names, [beta0_notched])
            logging.info("...done.\n")

        # covariance estimation window size, it may be modified by an internal flag in case of air-plane geometry
        cov_est_window_size = conf_params_obj.change_detection_fd.product_resolution

        if conf_params_obj.processing_flags.multilook_heading_correction:
            _, heading_deg, _, _, _, _ = decode_unique_acquisition_id_string(unique_stack_id + "_BSL_00")

            cov_est_window_size = resolution_heading_correction(cov_est_window_size, heading_deg)

        logging.info("Covariance Estimation Window Size used: {} [m]".format(cov_est_window_size))

        try:
            logging.info("FD: disturbance covariance matrix computation...")
            (MPMB_covariance_sr, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_covariance_estimation_SR(
                deepcopy(beta0_notched),
                cov_est_window_size,
                raster_info.pixel_spacing_slant_rg,
                raster_info.pixel_spacing_az,
                look_angle_rad,
                raster_info.carrier_frequency_hz,
                raster_info.range_bandwidth_hz,
            )

            del beta0_notched

            MPMB_covariance_sr = covariance_matrix_mat2vec(
                MPMB_covariance_sr
            )  # MPMB_covariance isreshaped from 3x3xNrgxNaz to 6xNrgxNaz

            # MPMB_covariance has 6 layers: saving each layer to numpy array and cancel the variable to free RAM space
            cov_layers_num, Nrg, Naz = MPMB_covariance_sr.shape  # cov_layers_num = 6
            MPMB_sr_temp_npy_fnames = []
            for layer_idx in np.arange(6):
                MPMB_sr_temp_npy_fnames.append(
                    os.path.join(temp_output_folder_sr, "MPMB_covariance_layer_{}_6.npy".format(layer_idx + 1))
                )
                np.save(MPMB_sr_temp_npy_fnames[layer_idx], MPMB_covariance_sr[layer_idx, :, :])

            # del MPMB_covariance_sr
            logging.info("...done.\n")

        except Exception as e:
            logging.error("FD: error during covariance computation " + str(e), exc_info=True)
            raise

        ### Interpolate it over a regular lat lon grid (with grid data): generate a regular grid for the interpolation by using Max and Min lat lon from the ECEFGRID_LLH (make the vectors a bit longer), for the grid steps use the minimum steps from the ECEFGRID_LLH)
        logging.info(unique_stack_id + ": Geocoding data...")
        try:

            # initialize the geocoding
            min_spacing_m = min(subs_F_a * raster_info.pixel_spacing_az, subs_F_r * raster_info.pixel_spacing_slant_rg)
            min_spacing_m = min(min_spacing_m, equi7_sampling)

            logging.info("Geocoding spacing set to {} [m]".format(min_spacing_m))

            (
                lon_in,
                lat_in,
                alt_in,
                lonMeshed_out,
                latMeshed_out,
                valid_values_mask,
                lon_regular_vector,
                used_lon_step,
                lat_regular_vector,
                used_lat_step,
            ) = geocoding_init(ecef_grid, rg_vec_subs, az_vec_subs, min_spacing_m)

            # geocode the MPMB_covariance
            logging.info(unique_stack_id + ": geocoding the covariance...")
            # geocode one layer at a time, loading from temporary numpy
            MPMB_gr_temp_npy_fnames = []
            for layer_idx in np.arange(cov_layers_num):
                curr_geocoded_fname = os.path.join(
                    temp_output_folder_gr, "MPMB_covariance_layer_{}_6.npy".format(layer_idx + 1)
                )

                # load slant range MPMB from npy file geocode and save the geocoded on  a new npy file
                MPMB_covariance_gr = geocoding(
                    np.load(MPMB_sr_temp_npy_fnames[layer_idx]),
                    lon_in,
                    lat_in,
                    lonMeshed_out,
                    latMeshed_out,
                    valid_values_mask,
                )
                np.save(curr_geocoded_fname, MPMB_covariance_gr)
                MPMB_gr_temp_npy_fnames.extend([curr_geocoded_fname])

            del MPMB_covariance_gr

            logging.info("...done.\n")

            # geocode the incidence angle
            logging.info(unique_stack_id + ": geocoding the incidence angle...")
            # incidence angle = off nadir angle - slope
            inc_angle_ground_rad = geocoding(
                (
                    off_nadir_angle_rad[master_id][rg_vec_subs, :][:, az_vec_subs]
                    - slope[rg_vec_subs, :][:, az_vec_subs]
                ),
                lon_in,
                lat_in,
                lonMeshed_out,
                latMeshed_out,
                valid_values_mask,
            )
            logging.info("...done.\n")

        except Exception as e:
            logging.error("FD: error during geocoding: " + str(e), exc_info=True)
            raise

        ### create GEOTIFF of all the layers (estimation, mask ):
        logging.info(unique_stack_id + ": formatting data to GEOTIFF...")

        try:

            cov_ground_fname = os.path.join(temp_output_folder_gr, "AverageCovariance.tif")
            inc_angle_ground_fname = os.path.join(temp_output_folder_gr, "IncidenceAngle.tif")

            upper_left_easting_coord = lon_regular_vector[0]  # i.e. horizontal
            sampling_step_east_west = used_lon_step
            upper_left_northing_coord = lat_regular_vector[0]  # i.e. vertical
            sampling_step_north_south = used_lat_step
            geotransform = [
                upper_left_easting_coord,
                sampling_step_east_west,
                0,
                upper_left_northing_coord,
                0,
                sampling_step_north_south,
            ]

            # FD covariance matrix geotiff formatting: six layers in the same geotiff (multilayer)
            tiff_formatter(
                MPMB_gr_temp_npy_fnames,
                cov_ground_fname,
                geotransform,
                gdal_data_format=gdal.GDT_CFloat32,
                multi_layers_tiff=True,
                time_tag=str(time_tag_mjd_curr),
            )

            # incidence angle matrix geotiff formatting:
            tiff_formatter(
                inc_angle_ground_rad,
                inc_angle_ground_fname,
                geotransform,
                gdal_data_format=gdal.GDT_Float32,
                multi_layers_tiff=False,
                time_tag=str(time_tag_mjd_curr),
            )
        except Exception as e:
            logging.error("FD: error during GEOTIFF formatting: " + str(e), exc_info=True)
            raise

        del inc_angle_ground_rad

        ### formatting data to EQUI7 and writing the XML LookUp Tables for fast indexing
        logging.info(unique_stack_id + ": formatting into EQUI7 grid...")
        try:

            equi7_COV_parent_tempdir = os.path.join(temp_output_folder_e7, "AverageCovariance")

            # in general from here the Equi7 can output multiple tiles, which file names are stored in the output list ( wrapped here in a dict for the stack )
            equi7_cov_temp_fnames = image2equi7grid(
                e7g,
                cov_ground_fname,
                equi7_COV_parent_tempdir,
                gdal_path=gdal_path,
                inband=None,
                subgrid_ids=None,
                accurate_boundary=False,
                withtilenamesuffix=False,
                resampling_type="bilinear",
                tile_nodata=np.nan,
            )

            equi7_inc_parent_tempdir = os.path.join(temp_output_folder_e7, "IncidenceAngle")
            equi7_inc_temp_fnames = image2equi7grid(
                e7g,
                inc_angle_ground_fname,
                equi7_inc_parent_tempdir,
                gdal_path=gdal_path,
                inband=None,
                subgrid_ids=None,
                accurate_boundary=False,
                withtilenamesuffix=False,
                resampling_type="bilinear",
                tile_nodata=np.nan,
            )

        except Exception as e:
            logging.error("FD: error during EQUI7 formatting: " + str(e), exc_info=True)
            raise

        if not os.path.exists(equi7_cov_temp_fnames[0]):
            error_message = "EQUI7 grid has not been generated, output is absent "
            logging.error(error_message)
            raise
        else:
            data_driver = gdal.Open(equi7_cov_temp_fnames[0], GA_ReadOnly)
            projection_equi7 = data_driver.GetProjection()
            data_driver = None

        logging.info("...done.\n")

        if time_step_idx == 0:
            recomputed_fnf_folder = os.path.join(temp_output_folder, "input_fnf", "equi7")
            if os.path.exists(recomputed_fnf_folder):
                # do not recompute fnf mask in output sampling if already present (because of precious cycles computation)
                equi7_initial_fnf_mask_fnames = get_equi7_tiff_names(recomputed_fnf_folder)

            else:
                # prepare the forest non forest mask
                # it is loaded if equi7, or converted to equi7 if TANDEM-X
                # it is a list containing all the loaded FNF-FTILES
                fnf_format = check_fnf_folder_format(
                    input_params_obj.stack_based_processing.forest_mask_catalogue_folder
                )
                if fnf_format == "TANDEM-X":
                    logging.info("Initial Forest mask is in TANDEM-X format, converting to equi7...")

                    # conversion step 1: get the tiff names of current zone
                    equi7_initial_fnf_mask_fnames = fnf_tandemx_load_filter_equi7format(
                        input_params_obj.stack_based_processing.forest_mask_catalogue_folder,
                        e7g,
                        conf_params_obj.change_detection_fd.product_resolution,
                        temp_output_folder,
                        gdal_path,
                        geographic_boundaries,
                        time_tag_mjd_initial,
                    )

                elif fnf_format == "EQUI7":
                    logging.info("Initial Forest mask reading and formatting...")

                    # re format the input mask in equi7 with the output resolution:
                    equi7_initial_fnf_mask_fnames = fnf_equi7_load_filter_equi7format(
                        input_params_obj.stack_based_processing.forest_mask_catalogue_folder,
                        e7g,
                        conf_params_obj.change_detection_fd.product_resolution,
                        temp_output_folder,
                        gdal_path,
                    )

        # check the equi7_fnf_mask_fnames
        check_equi7_mask_coverage(equi7_cov_temp_fnames, equi7_initial_fnf_mask_fnames)

        # equi7 zone is one contineltal zone
        equi7_zone_name = os.path.basename(os.path.dirname(os.path.dirname(equi7_cov_temp_fnames[0])))

        # current equi7 zone covariance product output folder
        equi7_zone_avg_cov_outdir = os.path.join(
            products_folder, "global_FD", "AverageCovariance", unique_stack_id, equi7_zone_name
        )
        equi7_zone_avg_cov_outdir_prev = os.path.join(
            products_folder, "global_FD", "AverageCovariance", unique_stack_id_prev, equi7_zone_name
        )

        # the disturbance computation is to be performed separately for each EQUI7 tile
        number_of_equi7_tiles = len(equi7_cov_temp_fnames)
        for equi7_tile_idx in np.arange(number_of_equi7_tiles):

            logging.info("FD: processing equi7 tile #{} of #{}".format(equi7_tile_idx + 1, number_of_equi7_tiles))

            # managing input/output paths and file names:
            equi7_tile_name = os.path.basename(
                os.path.dirname(equi7_cov_temp_fnames[equi7_tile_idx])
            )  # current equi7 tile name inside the equi7 zone

            # load the mask for the equi7 tile if not already present (calling it "final" because it will be updated at each global cycle)
            if not equi7_tile_name in final_forest_mask_data.keys():

                equi7_initial_fnf_mask_fname_curr = [
                    name for name in equi7_initial_fnf_mask_fnames if equi7_tile_name in name
                ][0]

                data_driver = gdal.Open(equi7_initial_fnf_mask_fname_curr, GA_ReadOnly)

                final_forest_mask_data[equi7_tile_name] = data_driver.GetRasterBand(1).ReadAsArray()

                time_tag_mjd_equi7_dict = data_driver.GetMetadata("TIFFTAG_DATETIME")
                data_driver = None
                if time_tag_mjd_equi7_dict:
                    time_tag_mjd_equi7 = PreciseDateTime().set_from_utc_string(time_tag_mjd_equi7_dict["time_tag"])

                    if time_tag_mjd_equi7 > time_tag_mjd_initial:
                        error_msg = 'Input FNF Mask (EQUI7) date of "{}" is bigger than input stack data minimum date of "{}"'.format(
                            str(time_tag_mjd_equi7), str(time_tag_mjd_initial)
                        )
                        logging.error(error_msg)
                        raise ValueError(error_msg)

            equi7_tile_avg_cov_outdir = os.path.join(
                equi7_zone_avg_cov_outdir, equi7_tile_name
            )  # full path of equi7 tile output folder

            equi7_cov_tiff_name_curr = "AverageCovariance.tif"  # current equi7 tiff file name inside the equi7 tile

            equi7_avg_cov_out_tiff_name_prev = os.path.join(
                equi7_zone_avg_cov_outdir_prev, equi7_tile_name, equi7_cov_tiff_name_curr
            )  # full path of equi7 tiff output file name at previous step (if present)

            if not os.path.exists(equi7_tile_avg_cov_outdir):
                os.makedirs(equi7_tile_avg_cov_outdir)

            # compure number of looks (matrix Nr)
            data_driver = gdal.Open(equi7_inc_temp_fnames[equi7_tile_idx], GA_ReadOnly)
            inc_angle_equi7_rad = data_driver.GetRasterBand(1).ReadAsArray()
            mat_number_of_looks = np.floor(
                (conf_params_obj.change_detection_fd.product_resolution ** 2)
                / (raster_info.resolution_m_az * raster_info.resolution_m_slant_rg / np.sin(inc_angle_equi7_rad))
            )
            np.save(os.path.join(temp_output_folder, "mat_number_of_looks.npy"), mat_number_of_looks)

            # at first iteration we should load the current covariance matrix (Yn)
            # if it is not present, use the first MPMB_covariance Xi as Yn = Xi and pass to next stack
            if time_step_idx == 0:
                prev_global_cycle_id = global_cycle_idx - 1
                prev_global_cycle_id_str = str(prev_global_cycle_id)

                equi7_avg_cov_out_tiff_name_prev = os.path.join(
                    input_params_obj.stack_based_processing.average_covariance_folder,
                    unique_stack_id_prev,
                    equi7_zone_name,
                    equi7_tile_name,
                    equi7_cov_tiff_name_curr,
                )

            if time_step_idx == 0 and not os.path.exists(equi7_avg_cov_out_tiff_name_prev):

                if prev_global_cycle_id < 0:
                    logging.info(
                        'FD first iteration, since first Global Cycle in input data is "0", cannot retrieve any previous Global Cycle Average Covariance'
                    )
                else:
                    logging.info(
                        "FD first iteration, input Average Covariance tiff not present at: {}".format(
                            equi7_avg_cov_out_tiff_name_prev
                        )
                    )

                logging.info(
                    "Saving current computed covariance as first Average Covariance (cannot compute disturbance at first iteration without input Average Covariance )"
                )

                # save the covariance to output and continue to next global cycle iteration
                shutil.move(equi7_cov_temp_fnames[equi7_tile_idx], equi7_tile_avg_cov_outdir)

                continue

            elif time_step_idx == 0 and os.path.exists(equi7_avg_cov_out_tiff_name_prev):

                logging.info(
                    "FD: first iteration, loading Average Covariance from of previous Global Cycle {}".format(
                        prev_global_cycle_id_str
                    )
                )

                # the starting Yn is an input:
                data_driver = gdal.Open(equi7_avg_cov_out_tiff_name_prev, GA_ReadOnly)
                for layer_idx in np.arange(6):
                    # saved matrix is 6xNrgxNaz, instead Y_N_matrix is 3x3xNrgxNaz matrix
                    if layer_idx == 0:
                        layer_read = data_driver.GetRasterBand(int(layer_idx + 1)).ReadAsArray()
                        Nrg_equi7, Naz_equi7 = layer_read.shape
                        Y_N_6x_vec = np.zeros((6, Nrg_equi7, Naz_equi7), dtype=np.complex64)
                        Y_N_6x_vec[layer_idx, :, :] = layer_read
                    else:
                        Y_N_6x_vec[layer_idx, :, :] = data_driver.GetRasterBand(int(layer_idx + 1)).ReadAsArray()

                data_driver = None

            logging.info('FD, global cycle "GC_' + global_cycle_idx_str + '" : computing distubance...')

            if time_step_idx != 0:
                # loading cumulate covariance at previous step:
                data_driver = gdal.Open(equi7_avg_cov_out_tiff_name_prev, GA_ReadOnly)

                for layer_idx in np.arange(6):

                    if layer_idx == 0:
                        layer_read = data_driver.GetRasterBand(int(layer_idx + 1)).ReadAsArray()
                        Nrg_equi7, Naz_equi7 = layer_read.shape
                        Y_N_6x_vec = np.zeros((6, Nrg_equi7, Naz_equi7), dtype=np.complex64)
                        Y_N_6x_vec[layer_idx, :, :] = layer_read
                    else:
                        Y_N_6x_vec[layer_idx, :, :] = data_driver.GetRasterBand(int(layer_idx + 1)).ReadAsArray()

                del layer_read
                data_driver = None

            # loading covariance at current step:
            data_driver = gdal.Open(equi7_cov_temp_fnames[equi7_tile_idx], GA_ReadOnly)

            for layer_idx in np.arange(6):

                if layer_idx == 0:
                    layer_read = data_driver.GetRasterBand(int(layer_idx + 1)).ReadAsArray()
                    X_i_6x_vec = np.zeros((6, Nrg_equi7, Naz_equi7), dtype=np.complex64)
                else:
                    X_i_6x_vec[layer_idx, :, :] = data_driver.GetRasterBand(int(layer_idx + 1)).ReadAsArray()

            del layer_read
            data_driver = None

            j_idx = 1
            disturbance_matrix = np.zeros((Nrg_equi7, Naz_equi7), dtype=np.bool)
            disturbance_prob_matrix = np.zeros((Nrg_equi7, Naz_equi7), dtype=np.bool)
            disturbance_pixel_validity = np.isnan(np.sum(X_i_6x_vec, 0))
            Nrg_equi7_str = str(Nrg_equi7)
            for rg_idx in np.arange(Nrg_equi7):

                if not np.remainder(rg_idx, 50):
                    logging.info("    change detection step {} of ".format(rg_idx) + Nrg_equi7_str)

                for az_idx in np.arange(Naz_equi7):

                    if disturbance_pixel_validity[rg_idx, az_idx]:
                        continue

                    X_i = covariance_matrix_vec2mat(X_i_6x_vec[:, rg_idx, az_idx])  # X_i is 3x3

                    Y_N = covariance_matrix_vec2mat(Y_N_6x_vec[:, rg_idx, az_idx])  # Y_N is 3x3
                    cov_number_of_looks = mat_number_of_looks[rg_idx, az_idx]

                    (
                        Y_N,
                        j_idx,
                        disturbance_matrix[rg_idx, az_idx],
                        disturbance_prob_matrix[rg_idx, az_idx],
                        _,
                    ) = alg_wishart_SU(
                        X_i,
                        Y_N,
                        j_idx,
                        number_of_pols,
                        cov_number_of_looks,
                        conf_params_obj.change_detection_fd.confidence_level,
                        False,
                    )

                    # update the matrix for the next stack
                    Y_N_6x_vec[:, rg_idx, az_idx] = covariance_matrix_mat2vec(Y_N)

            logging.info("...done.\n")

            logging.info("updating fnf mask for current global cycle")
            final_forest_mask_data[equi7_tile_name][disturbance_matrix] = False

            logging.info("saving updated average covariance for current global cycle")

            # SAVE Equi7 Average Covariance
            equi7_initial_fnf_mask_fname_curr = [
                name for name in equi7_initial_fnf_mask_fnames if equi7_tile_name in name
            ][0]
            data_driver = gdal.Open(equi7_initial_fnf_mask_fname_curr, GA_ReadOnly)
            geotransform_out = data_driver.GetGeoTransform()
            data_driver = None

            equi7_COV_out_fname = os.path.join(equi7_tile_avg_cov_outdir, equi7_cov_tiff_name_curr)
            tiff_formatter(
                [
                    Y_N_6x_vec[0, :, :],
                    Y_N_6x_vec[1, :, :],
                    Y_N_6x_vec[2, :, :],
                    Y_N_6x_vec[3, :, :],
                    Y_N_6x_vec[4, :, :],
                    Y_N_6x_vec[5, :, :],
                ],
                equi7_COV_out_fname,
                geotransform_out,
                gdal_data_format=gdal.GDT_Float32,
                projection=projection_equi7,
                multi_layers_tiff=True,
                time_tag=str(time_tag_mjd_curr),
            )

            logging.info(
                "FD: equi7 tile {} of {}".format(equi7_tile_idx, len(equi7_cov_temp_fnames))
                + " average covariange processing done.\n"
            )

        # keep track of the unique_stack_id for next iteration
        unique_stack_id_prev = unique_stack_id

        logging.info(
            'FD: disturbance computation for global cycle "GC_'
            + global_cycle_idx_str
            + '" of geometry '
            + nominal_geometry_stack_id
            + " done.\n"
        )

    ################## CYCLES LOOP END ##############################

    # save the final mask (one for each equi7 tile) and compute the dsturbance (one for each equi7 tile too)
    time_tag_mjd_final = time_tag_mjd_curr
    equi7_zone_disturbance_outdir = os.path.join(
        products_folder, "global_FD", "Disturbance", nominal_geometry_stack_id, equi7_zone_name
    )
    equi7_zone_fnf_mask_outdir = os.path.join(
        products_folder, "global_FD", "FNF_Mask", nominal_geometry_stack_id, equi7_zone_name
    )

    for equi7_tile_idx in np.arange(number_of_equi7_tiles):

        equi7_tile_name = os.path.basename(os.path.dirname(equi7_cov_temp_fnames[equi7_tile_idx]))

        equi7_tile_disturbance_outdir = os.path.join(equi7_zone_disturbance_outdir, equi7_tile_name)

        equi7_fnf_mask_tile_outdir = os.path.join(equi7_zone_fnf_mask_outdir, equi7_tile_name)
        equi7_disturbance_tiff_name = "disturbance.tif"
        equi7_fnf_mask_tiff_name = "fnf.tif"

        equi7_disturbance_out_fname = os.path.join(equi7_tile_disturbance_outdir, equi7_disturbance_tiff_name)
        equi7_fnf_mask_outfname = os.path.join(equi7_fnf_mask_tile_outdir, equi7_fnf_mask_tiff_name)

        if not os.path.exists(equi7_tile_disturbance_outdir):
            os.makedirs(equi7_tile_disturbance_outdir)
        if not os.path.exists(equi7_fnf_mask_tile_outdir):
            os.makedirs(equi7_fnf_mask_tile_outdir)

        logging.info(
            "FD: final step, computation of disturbance product, for equi7 tile #{} of #{}".format(
                equi7_tile_idx + 1, number_of_equi7_tiles
            )
        )

        equi7_initial_fnf_mask_fname_curr = [name for name in equi7_initial_fnf_mask_fnames if equi7_tile_name in name][
            0
        ]
        data_driver = gdal.Open(equi7_initial_fnf_mask_fname_curr, GA_ReadOnly)
        initial_forest_mask_curr = data_driver.GetRasterBand(1).ReadAsArray()
        geotransform_out = data_driver.GetGeoTransform()
        data_driver = None

        final_disturbance_matrix_curr = np.logical_xor(
            initial_forest_mask_curr, final_forest_mask_data[equi7_tile_name]
        )

        tiff_formatter(
            final_disturbance_matrix_curr,
            equi7_disturbance_out_fname,
            geotransform_out,
            gdal_data_format=gdal.GDT_Float32,
            projection=projection_equi7,
            time_tag=str(time_tag_mjd_final),
        )

        logging.info(
            "FD: final step,saving the final fnf mask for equi7 tile #{} of #{}".format(
                equi7_tile_idx, number_of_equi7_tiles
            )
        )
        tiff_formatter(
            final_forest_mask_data[equi7_tile_name],
            equi7_fnf_mask_outfname,
            geotransform_out,
            gdal_data_format=gdal.GDT_Float32,
            projection=projection_equi7,
            time_tag=str(time_tag_mjd_final),
        )

    # return the final_forest_mask_data to be updated in the next stack-cycle
    return final_forest_mask_data, temp_output_folder


def fill_core_processing_fd_obj(cycles_composition):

    """
    Internal function called by the StackBasedProcessingAGB APP:
        
        the InitializeAlgorithmFD APP fills the structure
        to be written into the xml input file for the next APP, which is the 
        core processing for the FD (change detection).
        
        Usage of the returned object:
            The returned object can be added to the input_params_obj and it 
            can be written to disk if needed, as (core_processing_fd_obj is 
                                                  overwritten by this command, if already present):
                - input_params_obj.core_processing_fd_obj = core_processing_fd_obj
                - write_input_file(input_params_obj, input_file_xml)
    """

    core_processing_fd_obj = core_processing_fd(cycles_composition)

    return core_processing_fd_obj



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import logging
import pyproj
import shutil
import numpy as np
from osgeo import gdal
from osgeo.gdalconst import GA_ReadOnly
from scipy.interpolate import interp1d
from skimage.filters.rank import majority as majority_filter
from equi7grid.equi7grid import Equi7Grid
from equi7grid.image2equi7grid import image2equi7grid
from biopal.io.data_io import (
    read_data,
    read_ecef_grid,
    read_auxiliary_multi_channels,
    read_auxiliary_single_channel,
    tandemx_fnf_read,
    tiff_formatter,
    readBiomassHeader,
)
from biopal.io.xml_io import raster_info
from biopal.utility.constants import OVERSAMPLING_FACTOR
from biopal.utility.utility_functions import get_equi7_fnf_tiff_names
from arepytools.io.productfolder import ProductFolder


def data_oversample(data, oversampling_factor, raster_info_obj):

    rg_ratio = np.floor(raster_info_obj.resolution_m_slant_rg / raster_info_obj.pixel_spacing_slant_rg)
    az_ratio = np.floor(raster_info_obj.resolution_m_az / raster_info_obj.pixel_spacing_az)

    rg_oversampling_flag = False
    az_oversampling_flag = False

    if rg_ratio < 2:
        rg_oversampling_flag = True
        num_samples_out = raster_info_obj.num_samples * oversampling_factor
    else:
        num_samples_out = raster_info_obj.num_samples
        pixel_spacing_slant_rg_out = raster_info_obj.pixel_spacing_slant_rg

    if az_ratio < 2:
        az_oversampling_flag = True
        num_lines_out = raster_info_obj.num_lines * oversampling_factor
    else:
        num_lines_out = raster_info_obj.num_lines
        pixel_spacing_az_out = raster_info_obj.pixel_spacing_az

    if rg_oversampling_flag or az_oversampling_flag:
        logging.info("    Oversampling needed:")

        if not type(data) is dict:
            # slope and reference_heights cases  (input data is value matrix)

            if rg_oversampling_flag:

                logging.info("        range   oversampling of auxiliary data...")

                data, pixel_spacing_slant_rg_out = data_oversample_core(
                    data, 0, raster_info_obj.pixel_spacing_slant_rg, oversampling_factor
                )

            if az_oversampling_flag:

                logging.info("        azimuth   oversampling of auxiliary data...")

                data, pixel_spacing_az_out = data_oversample_core(
                    data, 1, raster_info_obj.pixel_spacing_az, oversampling_factor
                )

        else:

            for data_key, data_extracted in data.items():

                if not type(data_extracted) is dict:

                    # KZ, off_nadir and ECEFGRID case (input data is a dict of values)
                    if rg_oversampling_flag:

                        logging.info("        range   oversampling of " + data_key + "...")

                        data[data_key], pixel_spacing_slant_rg_out = data_oversample_core(
                            data_extracted, 0, raster_info_obj.pixel_spacing_slant_rg, oversampling_factor,
                        )

                    if az_oversampling_flag:
                        logging.info("        azimuth oversampling of " + data_key + "...")

                        data_extracted = data[data_key]
                        data[data_key], pixel_spacing_az_out = data_oversample_core(
                            data_extracted, 1, raster_info_obj.pixel_spacing_az, oversampling_factor
                        )

                else:

                    # Beta0 case (input data is a dict of dict with values)
                    for pol_key, data_pol in data_extracted.items():

                        if rg_oversampling_flag:

                            logging.info(
                                "        range   oversampling of " + data_key + " , polarization " + pol_key + "..."
                            )

                            (data[data_key][pol_key], pixel_spacing_slant_rg_out,) = data_oversample_core(
                                data_pol, 0, raster_info_obj.pixel_spacing_slant_rg, oversampling_factor,
                            )

                        if az_oversampling_flag:
                            logging.info(
                                "        azimuth oversampling of " + data_key + " , polarization " + pol_key + "..."
                            )

                            data_pol = data[data_key][pol_key]
                            data[data_key][pol_key], pixel_spacing_az_out = data_oversample_core(
                                data_pol, 1, raster_info_obj.pixel_spacing_az, oversampling_factor
                            )

        logging.info("    oversampling done.\n")

    return data, num_samples_out, pixel_spacing_slant_rg_out, num_lines_out, pixel_spacing_az_out


def data_oversample_core(data, axis_index, pixel_spacing_in, oversampling_factor):

    # original size of the axis to be interpolated
    axis_len_in = data.shape[axis_index]

    # pixel spacing after the interpolation
    pixel_spacing_out = pixel_spacing_in / oversampling_factor

    # input original axis
    max_val_in = axis_len_in * pixel_spacing_in
    ax_in = np.arange(0, max_val_in, pixel_spacing_in)

    # output interpolated axis
    max_val_out = axis_len_in * oversampling_factor * pixel_spacing_out
    ax_out = np.arange(0, max_val_out, pixel_spacing_out)

    # interpolation of data along axis_index
    interp_fun = interp1d(ax_in, data, axis=axis_index, bounds_error=0)
    data = interp_fun(ax_out)

    return data, pixel_spacing_out


def read_and_oversample_data(L1c_repository, acquisitions_pf_names, enable_resampling):
    # this function calls the read_data followed by the data_oversample
    # (which oversamples only when needed and if enabled )

    beta0_calibrated = {}
    for pf_name in acquisitions_pf_names:
        logging.info("    loading " + pf_name + "...")
        # every data in the current stack has same spacing, same resolution and same number of samples and lines
        (
            beta0_calibrated[pf_name],
            num_samples,
            num_lines,
            pixel_spacing_slant_rg,
            pixel_spacing_az,
            carrier_frequency_hz,
            range_bandwidth_hz,
            master_id,
            lines_start_utc,
        ) = read_data(L1c_repository, pf_name)

    ### ALL chains: oversampling data:
    # needed whenever computing covariance or detected data (as notch for AGB)
    (_, _, _, _, _, _, resolution_m_slant_rg, resolution_m_az, sensor_velocity,) = readBiomassHeader(
        ProductFolder(os.path.join(L1c_repository, acquisitions_pf_names[0]), "r"), 0
    )

    # backup needed for the resampling in auxiliary data
    raster_info_orig = raster_info(
        num_samples,
        num_lines,
        pixel_spacing_slant_rg,
        pixel_spacing_az,
        resolution_m_slant_rg,
        resolution_m_az,
        carrier_frequency_hz,
        range_bandwidth_hz,
        lines_start_utc,
    )

    if enable_resampling:

        (beta0_calibrated, num_samples, pixel_spacing_slant_rg, num_lines, pixel_spacing_az,) = data_oversample(
            beta0_calibrated, OVERSAMPLING_FACTOR, raster_info_orig,
        )

        logging.info("all data loaded.\n")

    # filling output structure:
    raster_info_os = raster_info(
        num_samples,
        num_lines,
        pixel_spacing_slant_rg,
        pixel_spacing_az,
        resolution_m_slant_rg,
        resolution_m_az,
        carrier_frequency_hz,
        range_bandwidth_hz,
        lines_start_utc,
    )

    return beta0_calibrated, master_id, raster_info_os, raster_info_orig


def read_and_oversample_aux_data(
    file_names,
    stack_id,
    acquisitions_pf_names,
    enable_resampling,
    raster_info,
    read_ecef=True,
    read_off_nadir=True,
    read_slope=True,
    read_kz=True,
    read_ref_h=True,
    read_dist=True,
    read_average_cov=False,
    read_cal_screens=False,
    read_FH=False,
    read_reference_agb=False,
    read_sys_dec_fun=False,
):

    # this function calls the auxilary data readers followed by the data_oversample
    # (which oversamples only when needed and if enabled )
    # it reads only the auxiliary in the current stack_id

    # for TOMO we need another auxiliary: phases to be extreacted

    ecef_grid = None
    kz = None
    off_nadir_angle_rad = None
    reference_height = None
    R = None
    slope = None
    average_covariance = None
    calibration_screens = None
    cal_screens_raster_info = None
    forest_height = None
    reference_agb = None
    system_decorr_fun = None

    if read_ecef:
        logging.info("Loading auxiliary data: ECEFGRID...")
        pf_name = os.path.basename(file_names.ECEF_grid_file_names[stack_id])
        folder_name = os.path.dirname(file_names.ECEF_grid_file_names[stack_id])
        ecef_grid = read_ecef_grid(folder_name, pf_name)

        if not ecef_grid is None and enable_resampling:
            ecef_grid = data_oversample(ecef_grid, OVERSAMPLING_FACTOR, raster_info)[0]
        if not ecef_grid is None:
            logging.info("...ECEFGRID read.\n")
        else:
            logging.warning(
                'Since "ECEFGRID/'
                + stack_id
                + '" folder is missing into AuxiliaryProductsFolder, geometric library needs to be called.\n'
            )

    if read_kz:
        logging.info("Loading auxiliary data: KZ...")
        pf_name = os.path.basename(file_names.kz_file_names[stack_id])
        folder_name = os.path.dirname(file_names.kz_file_names[stack_id])
        kz = read_auxiliary_multi_channels(folder_name, pf_name, acquisitions_pf_names)

        if not kz is None and enable_resampling:
            kz = data_oversample(kz, OVERSAMPLING_FACTOR, raster_info)[0]
        if not kz is None:
            logging.info("...KZ read.\n")
        else:
            logging.warning(
                'Since "KZ/'
                + stack_id
                + '" folder is missing into AuxiliaryProductsFolder, geometric library needs to be called.\n'
            )

    if read_off_nadir:
        logging.info("Loading auxiliary data: off nadir angles...")
        pf_name = os.path.basename(file_names.off_nadir_angle_file_names[stack_id])
        folder_name = os.path.dirname(file_names.off_nadir_angle_file_names[stack_id])
        off_nadir_angle_rad = read_auxiliary_multi_channels(folder_name, pf_name)

        if not off_nadir_angle_rad is None and enable_resampling:
            off_nadir_angle_rad = data_oversample(off_nadir_angle_rad, OVERSAMPLING_FACTOR, raster_info)[0]
        if not off_nadir_angle_rad is None:
            logging.info("...off nadir angles read.\n")
        else:
            logging.warning(
                'Since "OffNadirAngles/'
                + stack_id
                + '" folder is missing into AuxiliaryProductsFolder, geometric library needs to be called.\n'
            )

    if read_ref_h:
        logging.info(
            "Loading auxiliary data: reference height..."
        )  # not needed, use just for comparison with the estimation
        pf_name = os.path.basename(file_names.reference_height_file_names[stack_id])
        folder_name = os.path.dirname(file_names.reference_height_file_names[stack_id])
        reference_height = read_auxiliary_single_channel(
            folder_name, pf_name
        )  # it is the dtm in slant_range-azimuth reference

        if not reference_height is None and enable_resampling:
            reference_height = data_oversample(reference_height, OVERSAMPLING_FACTOR, raster_info)[0]
        if not reference_height is None:
            logging.info("...reference height read.\n")
        else:
            logging.warning(
                'Since "ReferenceHeight/'
                + stack_id
                + '" folder is missing into AuxiliaryProductsFolder, geometric library needs to be called.\n'
            )

    if read_dist:
        logging.info("Loading auxiliary data: slant range distances...")
        pf_name = os.path.basename(file_names.slant_range_distances_file_names[stack_id])
        folder_name = os.path.dirname(file_names.slant_range_distances_file_names[stack_id])
        R = read_auxiliary_multi_channels(folder_name, pf_name)

        if not R is None and enable_resampling:
            R = data_oversample(R, OVERSAMPLING_FACTOR, raster_info)[0]
        if not R is None:
            logging.info("...slant range distances read.\n")
        else:
            logging.warning(
                'Since "SlantRangeDistances/'
                + stack_id
                + '" folder is missing into AuxiliaryProductsFolder, geometric library needs to be called.\n'
            )

    if read_slope:
        logging.info("Loading auxiliary data: slope...")
        pf_name = os.path.basename(file_names.slope_file_names[stack_id])
        folder_name = os.path.dirname(file_names.slope_file_names[stack_id])
        slope = read_auxiliary_single_channel(folder_name, pf_name)

        if not slope.any() is None and enable_resampling:
            slope = data_oversample(slope, OVERSAMPLING_FACTOR, raster_info)[0]
        if not slope is None:
            logging.info("...slope read.\n")
        else:
            logging.warning(
                'Since "Slopes/'
                + stack_id
                + '" folder is missing into AuxiliaryProductsFolder, geometric library needs to be called.\n'
            )

    if read_average_cov:
        logging.info("Loading auxiliary data: Average covariance matrix...")
        pf_name = os.path.basename(file_names.average_covariance_folder)
        folder = os.path.dirname(file_names.average_covariance_folder)
        average_covariance = read_auxiliary_multi_channels(folder, pf_name, acquisitions_pf_names)

        if not average_covariance is None and enable_resampling:
            average_covariance = data_oversample(average_covariance, OVERSAMPLING_FACTOR, raster_info)[0]
        if not average_covariance is None:
            logging.info("... Average covariance matrix read.\n")

    if read_cal_screens:
        logging.info("Loading auxiliary data: Calibration Screens...")
        pf_name = os.path.basename(file_names.calibration_screens_file_names[stack_id])
        folder = os.path.dirname(file_names.calibration_screens_file_names[stack_id])
        calibration_screens, cal_screens_raster_info = read_auxiliary_multi_channels(
            folder, pf_name, acquisitions_pf_names, read_raster_info=True
        )

        if not calibration_screens is None and enable_resampling:
            calibration_screens = data_oversample(calibration_screens, OVERSAMPLING_FACTOR, raster_info)[0]
        if not calibration_screens is None:
            logging.info("... Calibration Screens read.\n")

    if read_FH:
        logging.info("Loading auxiliary data: forest estimated height..")
        pf_name = os.path.basename(file_names.forest_height_folder)
        folder = os.path.dirname(file_names.forest_height_folder)
        forest_height = read_auxiliary_single_channel(folder, pf_name, acquisitions_pf_names)

        if not forest_height is None and enable_resampling:
            forest_height = data_oversample(forest_height, OVERSAMPLING_FACTOR, raster_info)[0]
        logging.info("...done.\n")
        if not forest_height is None:
            logging.info("...forest estimated height read.\n")

    if read_reference_agb:
        logging.info("Loading auxiliary data: reference agb..")
        pf_name = os.path.basename(file_names.reference_agb_folder)
        folder = os.path.dirname(file_names.reference_agb_folder)
        reference_agb = read_auxiliary_single_channel(folder, pf_name, acquisitions_pf_names)

        if not reference_agb is None and enable_resampling:
            reference_agb = data_oversample(reference_agb, OVERSAMPLING_FACTOR, raster_info)[0]
        if not reference_agb is None:
            logging.info("...reference agb read.\n")

    if read_sys_dec_fun:
        logging.info("Loading auxiliary data: system decorrelation function...")
        pf_name = os.path.basename(file_names.system_decorrelation_fun_folder)
        folder = os.path.dirname(file_names.system_decorrelation_fun_folder)
        system_decorr_fun = read_auxiliary_single_channel(folder, pf_name, acquisitions_pf_names)

        if not system_decorr_fun is None and enable_resampling:
            system_decorr_fun = data_oversample(system_decorr_fun, OVERSAMPLING_FACTOR, raster_info)[0]
        if not system_decorr_fun is None:
            logging.info("...system decorrelation function read.\n")

    return (
        ecef_grid,
        off_nadir_angle_rad,
        slope,
        kz,
        reference_height,
        R,
        average_covariance,
        calibration_screens,
        cal_screens_raster_info,
        forest_height,
        reference_agb,
        system_decorr_fun,
    )


def get_equi7_tiff_names(directory):

    equi7_fnames = []

    equi7_tile_names = os.listdir(directory)
    for equi7_tile_name in equi7_tile_names:
        equi7_subtile_names = os.listdir(os.path.join(directory, equi7_tile_name))
        for equi7_subtile_name in equi7_subtile_names:
            tiff_names = os.listdir(os.path.join(directory, equi7_tile_name, equi7_subtile_name))
            for tiff_name in tiff_names:
                equi7_fnames.append(os.path.join(directory, equi7_tile_name, equi7_subtile_name, tiff_name))

    return equi7_fnames


def fnf_tandemx_load_filter_equi7format(
    forest_mask_catalogue_folder,
    e7g,
    product_resolution,
    output_folder,
    gdal_path,
    geographic_boundaries,
    time_tag_mjd_initial,
    dummy_mask_flag=False,
):
    # this function loads from the fnf catalog folder, the times which matches the boundaries
    # the output is also casted to float to be ready for incapsulation within a geotiff

    fnf_tile_loaded_list, time_tag_mjd_tandemx, geotransform_list = tandemx_fnf_read(
        forest_mask_catalogue_folder, geographic_boundaries
    )

    # majority filter
    fnf_mask_list = []
    for idx_fnf, fnf in enumerate(fnf_tile_loaded_list):

        len_x, len_y = fnf.shape
        # two longitide and latitude points in middle of the fnf mask:

        # geotransform = [ lon_start, lon_step, 0, lat_start, 0, lat_step]
        lon0 = geotransform_list[idx_fnf][0] + geotransform_list[idx_fnf][1] * len_x / 2
        lon1 = lon0 + geotransform_list[idx_fnf][1]
        lat0 = geotransform_list[idx_fnf][3] + geotransform_list[idx_fnf][5] * len_x / 2
        lat1 = lat0 + geotransform_list[idx_fnf][5]

        geod = pyproj.Geod(ellps="WGS84")

        _, _, fnf_spacing_lon = geod.inv(lon0, lat0, lon1, lat0)
        _, _, fnf_spacing_lat = geod.inv(lon0, lat0, lon0, lat1)

        windtm_x = np.int(np.round(product_resolution / fnf_spacing_lon / 2) * 2 + 1)
        windtm_y = np.int(np.round(product_resolution / fnf_spacing_lat / 2) * 2 + 1)

        selem = np.ones((windtm_x, windtm_y), dtype=int)
        if not dummy_mask_flag:
            fnf_filtered = majority_filter(fnf.astype(int), selem)

            fnf_filtered = fnf_filtered.astype(fnf.dtype)
        else:
            fnf_filtered = np.ones(fnf.shape, dtype=fnf.dtype)

        # round the mask to quantize it because it is boolean: NOTE, do not cast to bool  otherwise the NaN will become True!
        fnf_filtered = np.round(fnf_filtered)

        fnf_mask_list.append(fnf_filtered)

    if time_tag_mjd_tandemx > time_tag_mjd_initial:
        error_msg = 'Input FNF Mask (TANDEM-X) date of "{}" is bigger than input stack data minimum date of "{}"'.format(
            str(time_tag_mjd_tandemx), str(time_tag_mjd_initial)
        )
        logging.error(error_msg)
        raise ValueError(error_msg)

    # conversion step 2: fnf mask geotiff formatting: one single layer geotiff for each fnf "tamdem-x tile" (they will be merged togheter after equi7 conversion )
    fnf_mask_ground_dir_name = os.path.join(output_folder, "fnf_ground")
    fnf_mask_ground_dir_names = tiff_formatter(
        fnf_mask_list,
        fnf_mask_ground_dir_name,
        geotransform_list,
        gdal_data_format=gdal.GDT_Float32,
        multi_layers_tiff=False,
        time_tag=str(time_tag_mjd_tandemx),
    )

    # conversion step 3: special case for the equi7 convertion of all the fnf masks found
    equi7_fnf_mask_tempdir_tiles_not_merged = (
        []
    )  # thos will contain all the fnf names in equi7: all fnf tiles, all equi7 tiles
    for idx, fnf_name_curr in enumerate(fnf_mask_ground_dir_names):
        equi7_fnf_mask_parent_tempdir = os.path.join(output_folder, "input_fnf_not_merged_T{}".format(idx))
        equi7_out_name_curr = image2equi7grid(
            e7g,
            fnf_name_curr,
            equi7_fnf_mask_parent_tempdir,
            gdal_path=gdal_path,
            inband=None,
            subgrid_ids=None,
            accurate_boundary=False,
            withtilenamesuffix=False,
            resampling_type="bilinear",
            tile_nodata=np.float(0),
        )
        equi7_fnf_mask_tempdir_tiles_not_merged.extend(equi7_out_name_curr)

    # conversion step 4: merging fnf tiles
    equi7_fnf_mask_parent_tempdir = os.path.join(output_folder, "input_fnf")
    # managing final mask creation:
    equi7_fnf_mask_fnames = fnf_equi7_masks_merging(
        equi7_fnf_mask_tempdir_tiles_not_merged, equi7_fnf_mask_parent_tempdir
    )

    logging.info("...fnf mask conversion into equi7 done.\n")

    return equi7_fnf_mask_fnames


def fnf_equi7_load_filter_equi7format(
    forest_mask_catalogue_folder, e7g_curr_chain, product_resolution, output_folder, gdal_path
):

    # get the names of input equi7 fnf masks
    equi7_fnf_mask_tiff_names_list, equi7_subgrid_name = get_equi7_fnf_tiff_names(forest_mask_catalogue_folder)

    subgrid_code = equi7_subgrid_name[6:8]

    if not equi7_fnf_mask_tiff_names_list:
        error_str = "Cannot find any Equi7 forest mask in input"
        logging.error(error_str)
        raise ValueError(error_str)

    # temporary folder for the input mask conversion
    equi7_fnf_mask_parent_tempdir = os.path.join(output_folder, "input_fnf")

    # output folder of the mask that will be used in the current chainflow:
    equi7_fnf_mask_parent_dir = os.path.join(equi7_fnf_mask_parent_tempdir, "equi7")

    if not os.path.exists(equi7_fnf_mask_parent_dir):
        os.makedirs(equi7_fnf_mask_parent_dir)
    if not os.path.exists(equi7_fnf_mask_parent_tempdir):
        os.makedirs(equi7_fnf_mask_parent_tempdir)

    equi7_fnf_mask_tiff_names_list_out = []

    # for each input Equi7 tile, load, filter and save with current equi7 grid sampling
    for idx, equi7_fnf_mask_tiff_name in enumerate(equi7_fnf_mask_tiff_names_list):

        # 1) filtering
        fnf_driver = gdal.Open(equi7_fnf_mask_tiff_name, GA_ReadOnly)
        fnf_data = fnf_driver.ReadAsArray()
        geotransform = fnf_driver.GetGeoTransform()
        projection_equi7 = fnf_driver.GetProjection()
        len_x, len_y = fnf_data.shape
        east_axis = geotransform[0] + geotransform[1] * np.arange(fnf_driver.RasterXSize)
        north_axis = geotransform[3] + geotransform[5] * np.arange(fnf_driver.RasterYSize)
        fnf_driver = None

        windtm_x = np.int(np.round(product_resolution / geotransform[1] / 2) * 2 + 1)
        windtm_y = np.int(np.round(product_resolution / abs(geotransform[5]) / 2) * 2 + 1)

        selem = np.ones((windtm_x, windtm_y), dtype=int)
        fnf_filtered = majority_filter(fnf_data.astype(int), selem)
        len_x, len_y = fnf_filtered.shape
        fnf_filtered = fnf_filtered.astype(fnf_data.dtype)
        # round the mask to quantize it because it is boolean: NOTE, do not cast to bool  otherwise the NaN will become True!
        fnf_filtered = np.round(fnf_filtered)

        # 2) intermediate saving to tiff
        intermediate_filtered_tiff_name = os.path.join(equi7_fnf_mask_parent_tempdir, "fnf_T{}.tif".format(idx))

        intermediate_filtered_tiff_name = tiff_formatter(
            fnf_filtered,
            intermediate_filtered_tiff_name,
            geotransform,
            gdal_data_format=gdal.GDT_Float32,
            projection=projection_equi7,
        )

        # create the equi7grid object with the input mask sampling
        equi7_sampling_mask_in = int(geotransform[1])
        try:
            e7g_in = Equi7Grid(equi7_sampling_mask_in)
        except Exception as e:
            logging.error(
                "Input FNF Equi7 mask geotransform is not in the correct format :{} ".format(subgrid_code) + str(e),
                exc_info=True,
            )
            raise

        # [(left, lower), (right, upper)]
        try:
            lon_min, lat_min = getattr(e7g_in, subgrid_code).xy2lonlat(min(east_axis), min(north_axis))
            lon_max, lat_max = getattr(e7g_in, subgrid_code).xy2lonlat(max(east_axis), max(north_axis))
        except Exception as e:
            logging.error(
                'Cannot recognize input FNF Equi7 mask "{}" sub-grid folder name :'.format(subgrid_code) + str(e),
                exc_info=True,
            )
            raise

        bbox = [(lon_min, lat_min), (lon_max, lat_max)]
        ftiles = e7g_curr_chain.search_tiles_in_roi(bbox=bbox)

        out_equi7_folder_done = image2equi7grid(
            e7g_curr_chain,
            intermediate_filtered_tiff_name,
            equi7_fnf_mask_parent_dir,
            gdal_path=gdal_path,
            ftiles=ftiles,
            accurate_boundary=False,
            withtilenamesuffix=False,
            tile_nodata=np.nan,
        )

        for idx, equi7_fnf_mask_tiff_name in enumerate(out_equi7_folder_done):

            fnf_driver = gdal.Open(equi7_fnf_mask_tiff_name, GA_ReadOnly)
            fnf_data = fnf_driver.ReadAsArray()
            fnf_driver = None

            if np.sum(np.isnan(fnf_data)) == fnf_data.shape[0] * fnf_data.shape[1]:
                shutil.rmtree(os.path.dirname(equi7_fnf_mask_tiff_name))
            else:
                equi7_fnf_mask_tiff_names_list_out.append(equi7_fnf_mask_tiff_name)

    if not equi7_fnf_mask_tiff_names_list_out:
        error_str = "Cannot find a valid input FNF Equi7 mask"
        logging.error(error_str)
        raise ImportError(error_str)

    return equi7_fnf_mask_tiff_names_list_out


def fnf_equi7_masks_merging(fnf_mask_equi7_fnames_whole, temp_output_folder):
    # there are many fnf mask tiles
    # for each fnf mask tile there are many Equi7 tiles
    # equi7 tiles may be grouped in different sub grids

    # retrive the name of tge "EQUI7 sub grid (there will be just one)
    equi7_subgrid_name = os.path.basename(os.path.dirname(os.path.dirname(fnf_mask_equi7_fnames_whole[0])))
    # also prepare the empty dictionary for equi7 tile names
    equi7_tile_names = []

    # retrive the names of all the used "EQUI7 tiles" and store in the equi7_tile_names: take the names from the generated EQUI7 folders for the fnf_mask_equi7
    for (
        fnf_tile_name
    ) in (
        fnf_mask_equi7_fnames_whole
    ):  # fnf_mask_equi7_fnames is a list of lists: fnf_mask_equi7_fnames[ list of fnf tiles ][ list of equi7 tiles ]
        curr_equi7_tile_name = os.path.basename(os.path.dirname(fnf_tile_name))
        if not curr_equi7_tile_name in equi7_tile_names:
            equi7_tile_names.append(curr_equi7_tile_name)

    merged_fnf_equi7_names = []
    # we are in the current EQUI7 sub-grid, cycle over all the
    for equi7_tile_name in equi7_tile_names:
        # search in all the fnf folders for same names
        out_fnf_equi7_name = os.path.join(
            temp_output_folder, "equi7_fnf_tiles_merged", equi7_subgrid_name, equi7_tile_name, "fnf.tif",
        )

        # search all the equi7 tile name in each fnf tile and sum togheter
        for fnf_tiff_full_path_curr in fnf_mask_equi7_fnames_whole:
            if equi7_tile_name in fnf_tiff_full_path_curr:

                input_image_driver = gdal.Open(fnf_tiff_full_path_curr, GA_ReadOnly)
                loaded_equi7_tile = input_image_driver.ReadAsArray()

                geotransform = input_image_driver.GetGeoTransform()
                projection = input_image_driver.GetProjection()

                # replace:
                loaded_equi7_tile = np.round(loaded_equi7_tile)
                loaded_equi7_tile[loaded_equi7_tile == 2] = 0
                loaded_equi7_tile[loaded_equi7_tile == 3] = 0

                # mean:
                if not "data_sum" in locals():
                    data_sum = loaded_equi7_tile
                else:
                    where_loaded_only_is_not_zero = np.logical_and(data_sum == 0, loaded_equi7_tile != 0)
                    where_both_are_not_zero = np.logical_and(data_sum != 0, loaded_equi7_tile != 0)
                    data_sum[where_loaded_only_is_not_zero] = loaded_equi7_tile[where_loaded_only_is_not_zero]
                    data_sum[where_both_are_not_zero] = (
                        loaded_equi7_tile[where_both_are_not_zero] + data_sum[where_both_are_not_zero]
                    ) / 2

        # round:
        data_sum.astype(float)

        # write the tiff as an equi7
        Nx, Ny = data_sum.shape

        driver = gdal.GetDriverByName("GTiff")
        os.makedirs(os.path.dirname(out_fnf_equi7_name))
        outdata = driver.Create(out_fnf_equi7_name, Ny, Nx, 1, gdal.GDT_Float32)
        outdata.SetGeoTransform(geotransform)
        outdata.SetProjection(projection)
        outdata.GetRasterBand(1).WriteArray(data_sum)
        outdata.FlushCache()  # saves to disk
        outdata = None

        del data_sum

        merged_fnf_equi7_names.append(out_fnf_equi7_name)

    return merged_fnf_equi7_names


def fnf_and_validity_masks_merging(merged_fnf_equi7_names, validity_mask_equi7_fnames, temp_output_folder, stack_id):

    final_mask_equi7_names = []

    # retrive the name of tge "EQUI7 sub grid (there will be just one)
    equi7_subgrid_name = os.path.basename(os.path.dirname(os.path.dirname(merged_fnf_equi7_names[0])))
    # also prepare the empty dictionary for equi7 tile names
    equi7_tile_names = []

    # retrive the names of all the used "EQUI7 tiles" and store in the equi7_tile_names: take the names from the generated EQUI7 folders for the fnf_mask_equi7
    for (
        fnf_equi7_name
    ) in (
        merged_fnf_equi7_names
    ):  # fnf_mask_equi7_fnames is a list of lists: fnf_mask_equi7_fnames[ list of fnf tiles ][ list of equi7 tiles ]
        curr_equi7_tile_name = os.path.basename(os.path.dirname(fnf_equi7_name))
        if not curr_equi7_tile_name in equi7_tile_names:
            equi7_tile_names.append(curr_equi7_tile_name)

    # search all the equi7 tile name in each fnf tile and sum togheter
    for equi7_tile_name in equi7_tile_names:
        fnf_mask_name = [name for name in merged_fnf_equi7_names if equi7_tile_name in name]
        validity_mask_name = [name for name in validity_mask_equi7_fnames if equi7_tile_name in name]
        # both should exist:
        if fnf_mask_name and validity_mask_name:

            fnf_mask_name = fnf_mask_name[0]
            validity_mask_name = validity_mask_name[0]
            out_final_equi7_name = os.path.join(
                temp_output_folder, "equi7_global_fnf", stack_id, equi7_subgrid_name, equi7_tile_name, "fnf.tif",
            )

            input_image_driver_fnf = gdal.Open(fnf_mask_name, GA_ReadOnly)
            loaded_equi7_fnf = input_image_driver_fnf.ReadAsArray()

            input_image_driver_val = gdal.Open(validity_mask_name, GA_ReadOnly)
            loaded_equi7_val = input_image_driver_val.ReadAsArray()

            final_mask = np.logical_and(loaded_equi7_fnf, loaded_equi7_val)

            final_mask.astype(int)

            # write the tiff as an equi7
            Nx, Ny = final_mask.shape

            geotransform = input_image_driver_fnf.GetGeoTransform()
            projection = input_image_driver_fnf.GetProjection()

            driver = gdal.GetDriverByName("GTiff")
            os.makedirs(os.path.dirname(out_final_equi7_name))
            outdata = driver.Create(out_final_equi7_name, Ny, Nx, 1, gdal.GDT_Float32)
            outdata.SetGeoTransform(geotransform)
            outdata.SetProjection(projection)
            outdata.GetRasterBand(1).WriteArray(final_mask)
            outdata.FlushCache()  ##saves to disk!!
            outdata = None

            final_mask_equi7_names.append(out_final_equi7_name)

    return final_mask_equi7_names


def mosaiking(equi7_main_full_folder, mosaiking_out_folder):
    # equi7_main_full_folder is the folder containing the temporary equi7 generated by the processor:
    # the final mosaiking will prodice the output folder, containing only Equi7 tiles, with no mention to stack IDs
    # this function uses listdir to retrive files and folder names from equi7_main_full_folder

    # procedure is the following:
    # first iteration is to retrive all the EQUI7 TILE names (as "E045N050T1" )
    # second iterates foe each of the above EQUI7 TILE name and does:
    #    inner cycle over all the stack folders to get the names of the tiff which matches the EQUI7 tile name
    #

    # each input in data_equi7_fnames is a tiff with two layers
    # first later is data
    data_layer_index = 1
    # second layer is quality
    quality_layer_index = 2

    # get the names of the stack or merged-stacks folder generated by the processor
    equi7_stack_folder_names = os.listdir(equi7_main_full_folder)

    # retrive the name of tge "EQUI7 sub grid (there will be just one)
    equi7_subgrid_folder_name = os.listdir(os.path.join(equi7_main_full_folder, equi7_stack_folder_names[0]))[0]

    ### first iteration: retrive all the EQUI7 TILE names (as "E045N050T1" )
    equi7_tile_names = []
    # retrive the names of all the used "EQUI7 tiles" and store in the equi7_tile_names
    for equi7_stack_folder_name in equi7_stack_folder_names:

        equi7_subgrid_full_folder = os.path.join(
            equi7_main_full_folder, equi7_stack_folder_name, equi7_subgrid_folder_name
        )

        equi7_tile_folder_names = os.listdir(equi7_subgrid_full_folder)
        for equi7_tile_name in equi7_tile_folder_names:
            if not equi7_tile_name in equi7_tile_names:
                equi7_tile_names.append(equi7_tile_name)

    ### second iteration
    out_mosaiked_equi7_tiff_names = []

    # cycle to write a modaiked output for each Equi7 tile name:
    for equi7_tile_name in equi7_tile_names:

        # inner cycle to get all the tiff names of the same equi7 ile:
        equi7_tile_tiff_full_names = []
        for equi7_stack_folder_name in equi7_stack_folder_names:

            equi7_subgrid_folder = os.path.join(
                equi7_main_full_folder, equi7_stack_folder_name, equi7_subgrid_folder_name
            )
            equi7_tile_folder_names = os.listdir(equi7_subgrid_folder)

            for equi7_tile_name_curr in equi7_tile_folder_names:

                if equi7_tile_name_curr == equi7_tile_name:
                    equi7_tile_tiff_name = os.listdir(os.path.join(equi7_subgrid_folder, equi7_tile_name))[0]
                    equi7_tile_tiff_full_names.append(
                        os.path.join(equi7_subgrid_folder, equi7_tile_name, equi7_tile_tiff_name)
                    )

        # now cycle over all the tiff data corresponding to same equi7 tile, amd performthe mosaicking (mean them together)
        out_tile_equi7_full_folder = os.path.join(mosaiking_out_folder, equi7_subgrid_folder_name, equi7_tile_name)
        for equi7_tiff_name in equi7_tile_tiff_full_names:

            # read current tiff to be mosaiked:
            input_image_driver = gdal.Open(equi7_tiff_name, GA_ReadOnly)
            loaded_equi7_tile_data = input_image_driver.GetRasterBand(data_layer_index).ReadAsArray()
            loaded_equi7_tile_quality = input_image_driver.GetRasterBand(quality_layer_index).ReadAsArray()

            # mosaiking (mean where both, use une where just one) it with previous data
            if not "out_mosaiked_data" in locals():
                out_mosaiked_data = loaded_equi7_tile_data
                out_mosaiked_quality = loaded_equi7_tile_quality
            else:
                where_loaded_only_is_not_zero = np.logical_and(out_mosaiked_data == 0, loaded_equi7_tile_data != 0)
                where_both_are_not_zero = np.logical_and(out_mosaiked_data != 0, loaded_equi7_tile_data != 0)

                out_mosaiked_data[where_loaded_only_is_not_zero] = loaded_equi7_tile_data[where_loaded_only_is_not_zero]
                out_mosaiked_data[where_both_are_not_zero] = (
                    loaded_equi7_tile_data[where_both_are_not_zero] + out_mosaiked_data[where_both_are_not_zero]
                ) / 2

                out_mosaiked_quality[where_loaded_only_is_not_zero] = loaded_equi7_tile_quality[
                    where_loaded_only_is_not_zero
                ]
                out_mosaiked_quality[where_both_are_not_zero] = (
                    loaded_equi7_tile_quality[where_both_are_not_zero] + out_mosaiked_quality[where_both_are_not_zero]
                ) / 2

        # write the output final mosaiked tiff of the current equi7 tile:
        out_mosaiked_equi7_name = os.path.join(out_tile_equi7_full_folder, "FH.tif",)

        geotransform = input_image_driver.GetGeoTransform()
        projection = input_image_driver.GetProjection()

        # write the tiff as an equi7
        Nx, Ny = out_mosaiked_data.shape

        driver = gdal.GetDriverByName("GTiff")
        os.makedirs(os.path.dirname(out_mosaiked_equi7_name))
        outdata = driver.Create(out_mosaiked_equi7_name, Ny, Nx, 2, gdal.GDT_Float32)
        outdata.SetGeoTransform(geotransform)
        outdata.SetProjection(projection)
        outdata.GetRasterBand(data_layer_index).WriteArray(out_mosaiked_data)
        outdata.GetRasterBand(quality_layer_index).WriteArray(out_mosaiked_quality)
        outdata.FlushCache()  # saves to disk
        outdata = None

        out_mosaiked_equi7_tiff_names.append(out_mosaiked_equi7_name)

        del out_mosaiked_data

    return out_mosaiked_equi7_tiff_names


def apply_dem_flattening(beta0_calibrated, kz_in, reference_height, master_id, raster_info):

    for pf_name in beta0_calibrated.keys():
        for pol_id in beta0_calibrated[pf_name]:
            beta0_calibrated[pf_name][pol_id] = np.multiply(
                beta0_calibrated[pf_name][pol_id], np.exp(-1j * np.multiply(kz_in[pf_name], reference_height)),
            )

    return beta0_calibrated



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import logging
import numpy as np
from scipy.stats import chi2


def get_rho_omega(j, p, n):
    # j is the number of covariance matrices being considered
    # p is the dimension of the covariance matrix (.ie 2 for dual pol, 3 for full pol)
    # n is the number of looks
    rho = 1 - (2 * p ** 2 - 1) * (1 + 1 / (j * (j - 1))) / (6 * p * n)
    omega2 = -(p ** 2) * (1 - 1 / rho) ** 2 / 4 + p ** 2 * (p ** 2 - 1) * (
        1 + (2 * j - 1) / (j ** 2 * (j - 1) ** 2)
    ) / (24 * n ** 2 * rho ** 2)
    return rho, omega2


def funR_wishart_SU(Y, X, j, n, p):

    # Here X is the next time point and Y is the sum of the previous X's after
    # the last change was detected. j is the total number of data being
    # considered ie. those which have contributed to Y plus one.

    lnR = n * (
        p * (j * np.log(j) - (j - 1) * np.log(j - 1))
        + (j - 1) * np.log(np.linalg.det(Y))
        + np.log(np.linalg.det(X))
        - j * np.log(np.linalg.det(Y + X))
    )
    lnR = np.real(lnR)
    return lnR


def alg_wishart_SU(X, Y, j, p, n, alpha, verbose):

    change = 0
    rho, omega2 = get_rho_omega(j + 1, p, n)
    testStat = -2 * rho * funR_wishart_SU(Y, X, j + 1, n, p)

    if testStat > 0:
        probR = 1 - (
            chi2.cdf(testStat, p ** 2) + omega2 * (chi2.cdf(testStat, p ** 2 + 4) - chi2.cdf(testStat, p ** 2))
        )
        if verbose:
            logging.info("Marginal Hypothesis %0.4f {} \n".format(probR))

        prob = 1 - probR
        if probR < alpha:
            # Change detected
            Y = X
            j = 1
            change = 1
        else:
            Y = Y + X
            j = j + 1
    else:
        prob = 0

    return Y, j, change, prob, testStat


def _apply_index_(a, idx, axis=-1):
    '''
    Funtion to apply argsort indexes to an array over an axis.
    Modified from:
        https://stackoverflow.com/questions/11253495/numpy-applying-argsort-to-an-array
    '''
    i = list(np.ogrid[[slice(x) for x in a.shape]])
    i[axis] = idx
    return a[tuple(i)]
    

def generalized_eigendecomp(A, B, eps=1e-6):
    """
    Computes the generalized eigendecomposition of the two covariance
    matrices A and B.
    
    :math:`\mathbf{A} \mathbf{v} = \lambda \mathbf{B} \mathbf{v}`
    
    This functions accepts numpy-like broadcasting.

    Parameters
    ----------
    A : ndarray, shape (..., N, N)
        Covariance matrix or matrices A to perform the generalized
        eigendecomposition
    B : ndarray, shape (..., N, N)
        Covariance matrix or matrices B to perform the generalized
        eigendecomposition
    eps : float, optional
        Preloading to apply to the main diagonal to avoid numerical errors
        when covariance matrices are semi-definite or zeros.
        The default value is 1e-6.
        It can be set to 0 to disable preloading.

    Returns
    -------
    W : ndarray, shape (..., N)
        Generalized eigenvalues (real) between A and B matrices
    V : ndarray, shape (..., N, N)
        Generalized eigenvectors between A and B matrices
    
    References
    ----------
    Alonso González, A., López Martínez, C., Papathanassiou, K., & Hajnsek, I.
    (2020). Polarimetric SAR time series change analysis over agricultural
    areas. IEEE transactions on geoscience and remote sensing, 58(10),
    7317-7330.

    """
    # Apply main diagonal preloading
    if eps > 0:
         A = A + eps * np.eye(A.shape[-1])
         B = B + eps * np.eye(A.shape[-1])
    # Perform generalized eigendecompsition
    W,V = np.linalg.eig(np.linalg.solve(A, B))
    # Note: eigenvalues/vectors are not sorted, we need to sort them
    ind = np.argsort(W.real, axis=-1)
    # Apply argsort to eigenvalues & eigenvectors
    W = _apply_index_(W.real, ind, axis=-1)
    V = _apply_index_(V, ind[...,np.newaxis,:], axis=-1)
    return W, V
    

def generate_pinc_pdec_RGBimages(W, V, p_min=1, p_max=10):
    """
    Generate polarimetric change increase and decrease Pauli RGB images.

    Parameters
    ----------
    W : ndarray, shape (..., N)
        Generalized eigenvalues (should be real).
    V : ndarray, shape (..., N, N)
        Generalized eigenvectors.
    p_min : float, optional
        Minimum value (in dB) for the scaling of the increasing/decreasing 
        Pauli RGB images. The default is 1dB.
    p_max : float, optional
        Maximum value (in dB) for the scaling of the increasing/decreasing 
        Pauli RGB images. The default is 10dB.

    Returns
    -------
    pinc : ndarray, shape (..., 3)
        Change representation corresponding to the Pauli RGB of the increasing
        polarization states within the range (p_min, p_max) in dB.
    pdec : ndarray, shape (..., 3)
        Change representation corresponding to the Pauli RGB of the decreasing
        polarization states within the range (p_min, p_max) in dB.
    
    References
    ----------
    Alonso González, A., López Martínez, C., Papathanassiou, K., & Hajnsek, I.
    (2020). Polarimetric SAR time series change analysis over agricultural
    areas. IEEE transactions on geoscience and remote sensing, 58(10),
    7317-7330.

    """
    # Get generalized eigenvalues in dB
    Wl = 10*np.log10(W)
    # Compute pinc and pdec vectors
    pinc = np.linalg.norm(np.abs(V) * ((Wl * (Wl > 0))[..., np.newaxis,: ]),
                          axis=-1)
    pdec = np.linalg.norm(np.abs(V) * ((-Wl * (Wl < 0))[..., np.newaxis, :]),
                          axis=-1)
    # NOTE: The channel ordering has to be changed to match the classical
    # Pauli RGB representation
    pinc = pinc[...,(1,2,0)]
    pdec = pdec[...,(1,2,0)]
    # Clip images to the given change range
    pinc = (np.clip(pinc, p_min,p_max)-p_min)/(p_max-p_min)
    pdec = (np.clip(pdec, p_min,p_max)-p_min)/(p_max-p_min)
    return pinc, pdec
    


"""
biopal quick start

Usage:
  biopal-quickstart FOLDER
  
Arguments:
  FOLDER       Path to the folder to initialize
  
Options:
  -h --help          Show this screen
  --version          Show version
  
Details:
    Initialize the folder with default input and configuration xml files:
    Input_File.xml to be edited before BioPAL run
    Configuration_File.xml optionally to be edited before BioPAL run
"""

import sys
import pkgutil
from biopal import __version__
from pathlib import Path


def main():
    from docopt import docopt

    args = docopt(__doc__, version=__version__)

    folder = Path(args["FOLDER"])

    if folder.exists():
        print("Error: provided folder already exists", file=sys.stderr)
        sys.exit(1)

    folder.mkdir(parents=True)
    folder.joinpath("Input_File.xml").write_bytes(pkgutil.get_data("biopal", "_package_data/inputs/Input_File.xml"))
    folder.joinpath("Configuration_File.xml").write_bytes(
        pkgutil.get_data("biopal", "_package_data/conf/Configuration_File.xml")
    )
    print("Folder intialized '{}'".format(folder.absolute()))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import logging
import numpy as np
import pyproj as proj
from shutil import which
from datetime import datetime
from arepytools.timing.precisedatetime import PreciseDateTime
from arepytools.io.productfolder import ProductFolder
from arepytools.io.metadata import RasterInfo
from equi7grid.equi7grid import Equi7Grid
from biopal.utility.constants import LIGHTSPEED

# The class Task is the template to be inherited for the creation of each BioPAL APP
class Task:
    def __init__(self, configuration_file=None):

        self.configuration_file = configuration_file

    def name(self):
        return self.__class__.__name__

    def _run(self, input_file_xml):
        raise NotImplementedError()

    def run(self, input_file_xml):

        try:

            logging.info("Starting {}".format(self.name()))
            output = self._run(input_file_xml)
            logging.info("Finished {}".format(self.name()))

            return output

        except Exception as e:
            error_msg = "biopal error inside {}: {}".format(self.name(), e)
            logging.error(error_msg, exc_info=True)
            raise RuntimeError(error_msg)


def start_logging(output_folder, L2_product, log_level, app_name=None):
    # CRITICAL 50
    # ERROR 40
    # WARNING 30
    # INFO 20
    # DEBUG 10
    # NOTSET 0

    if log_level == "DEBUG":
        level_to_set = logging.DEBUG
    elif log_level == "INFO":
        level_to_set = logging.INFO
    elif log_level == "WARNING":
        level_to_set = logging.WARNING
    elif log_level == "ERROR":
        level_to_set = logging.ERROR

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    if not app_name:
        log_file_name = os.path.join(output_folder, "biomassL2.log")
    else:
        log_file_name = os.path.join(output_folder, app_name + "_APP.log")

    if os.path.exists(log_file_name):
        raise RuntimeError("output folder {} is not empty.".format(output_folder))

    logging.basicConfig(
        handlers=[logging.FileHandler(log_file_name, mode="w", encoding="utf-8"), logging.StreamHandler(),],
        level=level_to_set,
        format="%(asctime)s - %(levelname)s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    logging.getLogger("matplotlib.font_manager").disabled = True

    logging.info(" --BIOMASS L2 Processor-- ")
    if not app_name:
        logging.info("Executing BioPAL full processor, L2 product: {}".format(L2_product))
    else:
        logging.info("Executing {} APP".format(app_name))

    logging.info(" \n")

    return log_file_name


def set_gdal_paths(gdal_path, gdal_environment_path=None):
    if gdal_path is None or not gdal_path:
        gdal_info_path = which("gdalinfo")
        if not gdal_info_path:
            raise RuntimeError("Missing gdalinfo executable")
        gdal_path = os.path.dirname(gdal_info_path)

    if gdal_environment_path is None or not gdal_environment_path:
        gdal_environment_path = os.environ.get("GDAL_DATA")
        if not gdal_environment_path:
            raise RuntimeError("Missing GDAL_DATA environment variable")
    else:
        # Set the enviroment
        os.environ["GDAL_DATA"] = gdal_environment_path

    logging.info("gdal_path: " + gdal_path)
    logging.info("gdal_environment_path: " + gdal_environment_path)

    return gdal_path, gdal_environment_path


def get_data_time_stamp(folder, pf_name):
    # reads a data:
    # it is supposed to contain one or more polarizations (the "SwathInfo" is read to retrive it)
    # it returns a dictionary with keys = polarizations
    # it returns also the dimensions of data

    data_pf_name = os.path.join(folder, pf_name)

    pf = ProductFolder(data_pf_name, "r")

    # prepare the metadata elements
    data_channel_obj = pf.get_channel(0)
    metadata_obj = data_channel_obj.metadata
    metadatachannel_obj = metadata_obj.get_metadata_channels(0)

    # Raster Info
    ri = metadatachannel_obj.get_element("RasterInfo")

    lines_start_utc = str(ri.lines_start)

    return lines_start_utc


def getBinaryNameFromChannelIDX(pf_name, channel_idx):

    if not type(channel_idx) == type(1) or channel_idx < 1:
        error_message = "Input channel_idx should be an integer number greater than zero"
        logging.error(error_message)
        raise ValueError(error_message)

    channel_idx_str = str(channel_idx)
    number_of_zeros = 4 - len(channel_idx_str)
    zeros_str = "0" * number_of_zeros
    binary_name = pf_name + "_" + zeros_str + channel_idx_str

    return binary_name


def decode_unique_acquisition_id_string(unique_acquisition_id_string, output_format="numeric"):

    if output_format == "string":
        global_cycle_idx = unique_acquisition_id_string[3:5]
        heading_deg = unique_acquisition_id_string[8:14]
        rg_swath_idx = unique_acquisition_id_string[20:22]
        rg_sub_swath_idx = unique_acquisition_id_string[30:32]
        az_swath_idx = unique_acquisition_id_string[38:40]
        baseline_idx = unique_acquisition_id_string[45:]

    elif output_format == "numeric":
        global_cycle_idx = int(unique_acquisition_id_string[3:5])
        heading_deg = float(unique_acquisition_id_string[8:14])
        rg_swath_idx = int(unique_acquisition_id_string[20:22])
        rg_sub_swath_idx = int(unique_acquisition_id_string[30:32])
        az_swath_idx = int(unique_acquisition_id_string[38:40])
        baseline_idx = int(unique_acquisition_id_string[45:])

    else:
        raise ValueError(
            'Input output_format = "{}"  not valid, it should be "numeric" or "string"'.format(output_format)
        )

    return (
        global_cycle_idx,
        heading_deg,
        rg_swath_idx,
        rg_sub_swath_idx,
        az_swath_idx,
        baseline_idx,
    )


def check_if_path_exists(path, file_folder_str):
    if not os.path.exists(path):
        log_error_str = file_folder_str + " " + path + " does not exist"

        logging.error(log_error_str)
        raise Exception(log_error_str)


def check_if_geometry_auxiliaries_are_present(
    file_names,
    stack_id,
    acquisitions_pf_names,
    read_ecef=True,
    read_off_nadir=True,
    read_slope=True,
    read_kz=True,
    read_ref_h=True,
    read_dist=True,
):

    aux_are_present_flag = True

    if read_ecef:
        pf_name = os.path.basename(file_names.ECEF_grid_file_names[stack_id])
        folder_name = os.path.dirname(file_names.ECEF_grid_file_names[stack_id])
        if not os.path.exists(os.path.join(folder_name, pf_name)):
            aux_are_present_flag = False

    if read_kz:
        pf_name = os.path.basename(file_names.kz_file_names[stack_id])
        folder_name = os.path.dirname(file_names.kz_file_names[stack_id])
        if not os.path.exists(os.path.join(folder_name, pf_name)):
            aux_are_present_flag = False

    if read_off_nadir:
        pf_name = os.path.basename(file_names.off_nadir_angle_file_names[stack_id])
        folder_name = os.path.dirname(file_names.off_nadir_angle_file_names[stack_id])
        if not os.path.exists(os.path.join(folder_name, pf_name)):
            aux_are_present_flag = False

    if read_ref_h:
        pf_name = os.path.basename(file_names.reference_height_file_names[stack_id])
        folder_name = os.path.dirname(file_names.reference_height_file_names[stack_id])
        if not os.path.exists(os.path.join(folder_name, pf_name)):
            aux_are_present_flag = False

    if read_dist:
        pf_name = os.path.basename(file_names.slant_range_distances_file_names[stack_id])
        folder_name = os.path.dirname(file_names.slant_range_distances_file_names[stack_id])
        if not os.path.exists(os.path.join(folder_name, pf_name)):
            aux_are_present_flag = False

    if read_slope:
        pf_name = os.path.basename(file_names.slope_file_names[stack_id])
        folder_name = os.path.dirname(file_names.slope_file_names[stack_id])
        if not os.path.exists(os.path.join(folder_name, pf_name)):
            aux_are_present_flag = False

    return aux_are_present_flag


def choose_equi7_sampling(product_resolution, geographic_grid_sampling):

    if geographic_grid_sampling <= product_resolution / 2:
        equi7_sampling = geographic_grid_sampling
    else:
        equi7_sampling = product_resolution / 2
        warning_message = "user rquested Geographic Grid Sampling is {} [m]; it cannot be greater than (product_resolution)/2 = {} [m] \n".format(
            geographic_grid_sampling, product_resolution / 2
        )
        logging.warning(warning_message)

    _static_sampling_sorted = np.sort(Equi7Grid._static_sampling)
    indexes_less_or_equal = np.where(np.less_equal(_static_sampling_sorted, equi7_sampling))

    if len(indexes_less_or_equal[0]):
        equi7_sampling_quantized = _static_sampling_sorted[indexes_less_or_equal[0][-1]]
    else:
        equi7_sampling_quantized = _static_sampling_sorted[0]

    return equi7_sampling_quantized


def evaluate_estimation_quality_matrix(data_in_shape):
    # Placemark for the quality estimation to be defined

    logging.warning(
        "The quality estimation of the product is still to be defined: zeros will be placed in the quality layer "
    )
    quality_matrix = np.zeros(data_in_shape)

    return quality_matrix


def epsg_in_to_epsg_out(xx, yy, zz, epsg_code_in, epsg_code_out):
    p_in = proj.Proj("+init={}".format(epsg_code_in))
    p_out = proj.Proj("+init={}".format(epsg_code_out))
    xx_transf, yy_transf, zz_transf = proj.transform(p_in, p_out, xx, yy, zz)

    return xx_transf, yy_transf, zz_transf


def convert_rasterinfo_meters_to_seconds(ri_meters, sensor_velocity):

    ri_seconds = RasterInfo(
        ri_meters.lines,
        ri_meters.samples,
        ri_meters.cell_type,
        ri_meters.file_name,
        ri_meters.header_offset_bytes,
        ri_meters.row_prefix_bytes,
        ri_meters.byte_order,
    )

    lines_step_s = ri_meters.lines_step / sensor_velocity
    lines_step_unit = "s"
    samples_start_s = ri_meters.samples_start / LIGHTSPEED * 2
    samples_start_unit = "s"
    samples_step_s = ri_meters.samples_step / LIGHTSPEED * 2
    samples_step_unit = "s"

    ri_seconds.set_lines_axis(ri_meters.lines_start, ri_meters.lines_start_unit, lines_step_s, lines_step_unit)
    ri_seconds.set_samples_axis(samples_start_s, samples_start_unit, samples_step_s, samples_step_unit)

    return ri_seconds


def convert_rasterinfo_seconds_to_meters(ri_seconds, sensor_velocity):

    ri_meters = RasterInfo(
        ri_seconds.lines,
        ri_seconds.samples,
        ri_seconds.cell_type,
        ri_seconds.file_name,
        ri_seconds.header_offset_bytes,
        ri_seconds.row_prefix_bytes,
        ri_seconds.byte_order,
    )

    lines_step_m = ri_meters.lines_step * sensor_velocity
    lines_step_unit = "m"
    samples_start_m = ri_meters.samples_start * LIGHTSPEED / 2
    samples_start_unit = "m"
    samples_step_m = ri_meters.samples_step * LIGHTSPEED / 2
    samples_step_unit = "m"

    ri_meters.set_lines_axis(
        ri_seconds.lines_start, ri_seconds.lines_start_unit, lines_step_m, lines_step_unit,
    )
    ri_meters.set_samples_axis(samples_start_m, samples_start_unit, samples_step_m, samples_step_unit)

    return ri_meters


def save_breakpoints(output_folder, breakpoint_names_list, breakpoint_data_list):

    if len(breakpoint_names_list) != len(breakpoint_data_list):
        raise IndexError("Inputs should have same shape")

    for idx, file_name in enumerate(breakpoint_names_list):
        np.save(os.path.join(output_folder, file_name), breakpoint_data_list[idx])


def format_folder_name():
    # date string format is: BIOMASS_L2_YYYYMMDDTHHMMSS
    now = datetime.now()
    year_str = str(now.year)
    month_str = str(now.month).rjust(2, "0")
    day_str = str(now.day).rjust(2, "0")
    h_str = str(now.hour).rjust(2, "0")
    m_str = str(now.minute).rjust(2, "0")
    s_str = str(now.second).rjust(2, "0")
    current_date_string = "BIOMASS_L2_" + year_str + month_str + day_str + "T" + h_str + m_str + s_str

    return current_date_string


def check_fnf_folder_format(fnf_mask_catalogue_folder):

    fnf_content_folders = os.listdir(fnf_mask_catalogue_folder)
    if not fnf_content_folders:
        error_msg = "input fnf mask is mandatory"
        logging.error(error_msg)
        raise Exception(error_msg)

    found_tandemx = False
    found_equi7 = False
    for fnf_content_folder in fnf_content_folders:
        if "TDM_FNF_" in fnf_content_folder:
            found_tandemx = True

        elif "EQUI7_" in fnf_content_folder:
            found_equi7 = True

        else:
            error_msg = 'Cannot recognize input FNF mask format: valid formats are EQUI7 (an "EQUI7_*/" folder containing equi7 tiles sub-folders, example: "EQUI7_AF025M/E042N048T3) or TANDEM-X (collection of "TDM_FNF_*/" folders '
            logging.error(error_msg)
            raise Exception(error_msg)

    if found_tandemx and found_equi7:
        error_msg = " Found bowth TANDEM-X and EQUI7 format FNF masks, this is not supported."
        logging.error(error_msg)
        raise Exception(error_msg)
    elif found_tandemx:
        return "TANDEM-X"
    elif found_equi7:
        return "EQUI7"


def check_cal_format(reference_agb_folder):

    cal_format = "RASTER"

    return cal_format


def get_min_time_stamp_repository(L1c_repository, stack_composition):

    for idx, (unique_stack_id, unique_acq_pf_names) in enumerate(stack_composition.items()):

        if idx == 0:
            time_tag_mjd_initial = get_min_time_stamp_repository_core(L1c_repository, unique_acq_pf_names)
        else:
            time_tag_mjd_initial = min(
                time_tag_mjd_initial, get_min_time_stamp_repository_core(L1c_repository, unique_acq_pf_names),
            )

    time_tag_mjd_initial = PreciseDateTime().set_from_utc_string(time_tag_mjd_initial)

    return time_tag_mjd_initial


def get_min_time_stamp_repository_core(L1c_repository, acquisitions_pf_names):
    for idx, pf_name in enumerate(acquisitions_pf_names):

        if idx == 0:
            min_time_stamp_mjd = get_data_time_stamp(L1c_repository, pf_name)
        else:
            min_time_stamp_mjd = min(min_time_stamp_mjd, get_data_time_stamp(L1c_repository, pf_name))

    return min_time_stamp_mjd


def check_equi7_mask_coverage(data_equi7_fnames, equi7_fnf_mask_fnames):
    equi7_tile_names = []
    for fname_curr in data_equi7_fnames:
        equi7_tile_names.append(os.path.basename(os.path.dirname(fname_curr)))
    input_mask_equi7_tile_names = []
    for m_name_curr in equi7_fnf_mask_fnames:
        input_mask_equi7_tile_names.append(os.path.basename(os.path.dirname(m_name_curr)))
    for equi7_tile_name in equi7_tile_names:
        if not equi7_tile_name in input_mask_equi7_tile_names:
            error_msg = 'The auxiliary "ForestMask" does not cover the acquisitions, missing Equi7 tile "{}"'.format(
                equi7_tile_name
            )
            logging.error(error_msg)
            raise ValueError(error_msg)


def resolution_heading_correction(resolution, heading_deg):

    if heading_deg < 0 or heading_deg > 360:
        error_str = "Stack heading not valid: {}  [deg]".format(heading_deg)
        logging.error(error_str)
        raise ValueError(error_str)

    if heading_deg >= 0 and heading_deg <= 45:
        correction_factor = np.cos(np.deg2rad(heading_deg))

    elif heading_deg > 45 and heading_deg <= 135:
        correction_factor = np.sin(np.deg2rad(heading_deg))

    elif heading_deg > 135 and heading_deg <= 225:
        correction_factor = np.abs(np.cos(np.deg2rad(heading_deg)))

    elif heading_deg > 225 and heading_deg <= 315:
        correction_factor = np.abs(np.sin(np.deg2rad(heading_deg)))

    elif heading_deg > 315:
        correction_factor = np.cos(np.deg2rad(heading_deg))

    resolution = resolution / correction_factor

    return resolution


def get_equi7_fnf_tiff_names(fnf_mask_catalogue_folder):

    equi7_tiles_subfolders = []
    e7g_type = ""
    for idx, equi7_subgrid_name in enumerate(os.listdir(fnf_mask_catalogue_folder)):

        if idx == 1:
            logging.warning(
                "Found more than one sub-grid in input Equi7 format FNF mask: prototype supports just one grid at a time, only {} sub-grid will be read".format(
                    equi7_subgrid_name
                )
            )
            break

        equi7_folder = os.path.join(fnf_mask_catalogue_folder, equi7_subgrid_name)
        temp_list = os.listdir(equi7_folder)
        if e7g_type == "":
            e7g_type = temp_list[0][-2:]
        if e7g_type in temp_list[0][-2:]:
            for folder_curr in temp_list:
                equi7_tiles_subfolders.append(os.path.join(equi7_folder, folder_curr))
        elif not e7g_type == temp_list[0][-2:]:
            logging.warning(
                'Auxiliary forest mask multiple tiles sampling are not supported, only "'
                + e7g_type
                + '" tile sampling have been read, "'
                + temp_list[0][-2:]
                + '" skipped instead'
            )

    equi7_fnf_mask_tiff_names_list = []
    for equi7_tile_folder in equi7_tiles_subfolders:

        tiff_fname = os.listdir(equi7_tile_folder)[0]

        equi7_fnf_mask_tiff_names_list.append(os.path.join(equi7_tile_folder, tiff_fname))

    return equi7_fnf_mask_tiff_names_list, equi7_subgrid_name


def get_raster_cal_names(reference_agb_folder):

    tiff_names = os.listdir(reference_agb_folder)

    cal_names = []
    for tiff_name in tiff_names:
        if ".tif" in tiff_name or ".tiff" in tiff_name:
            cal_names.append(os.path.join(reference_agb_folder, tiff_name))

    return cal_names


def get_foss_cal_names(reference_agb_folder):

    tiff_names = os.listdir(reference_agb_folder)

    cal_names = []
    for tiff_name in tiff_names:
        if ".tif" in tiff_name or ".tiff" in tiff_name:
            cal_names.append(os.path.join(reference_agb_folder, tiff_name))

    return cal_names


def collect_stacks_to_be_merged(stack_composition):
    """
    The stack_composition contains the IDs of all the stacks and
    for each of the stacks contains all the acquisition IDs.

    This function collects togeter all the stacks which have all in common except the heading:
    the resulting stacks_to_merge_dict is a dictionary containing
        as keys,   all the unique_merged_stack_id (stack IDs with missing headings)
        as values, the list of the stack_ids to be merged (same stacks, all headings)

    """
    stacks_to_merge_dict = {}
    for unique_stack_id in stack_composition.keys():

        # 1) read the acquisition ID of first acquisition in the current stack
        (
            global_cycle_idx,
            heading_deg,
            rg_swath_idx,
            rg_sub_swath_idx,
            az_swath_idx,
            baseline_idx,
        ) = decode_unique_acquisition_id_string(stack_composition[unique_stack_id][0], output_format="string")

        # 2) create a new ID which is a stack_id without the heading
        unique_merged_stack_id = (
            "GC_" + global_cycle_idx + "_RGSW_" + rg_swath_idx + "_RGSBSW_" + rg_sub_swath_idx + "_AZSW_" + az_swath_idx
        )

        if unique_merged_stack_id in stacks_to_merge_dict.keys():
            stacks_to_merge_dict[unique_merged_stack_id].append(unique_stack_id)
        else:
            stacks_to_merge_dict[unique_merged_stack_id] = [unique_stack_id]

    return stacks_to_merge_dict




import nox
from pathlib import Path

"""
Run from command line functionalities in isolated python virtual environment

Usage:
    nox -r -fb conda -s FUNCTION
see also nox --help

Details:
    available functions are:
    build_sdist: sdist package for PyPI 
    build_wheel: wheel package for PyPI
    build_doc: generate sphinx documentation from docstrings and rst files
"""


@nox.session()
def build_sdist(session: nox.Session):
    session.install("build")
    session.run(
        "python", "-m", "build", "--sdist", silent=True,
    )


@nox.session()
def build_wheel(session: nox.Session):
    session.install("build")
    session.run(
        "python", "-m", "build", "--wheel", silent=True,
    )


@nox.session(python="3.7.1")
def build_doc(session: nox.Session):
    session.conda_install("GDAL")
    session.install("-e", ".")
    session.install("-r", "doc/requirements.txt")
    rst_files = [
        "chains/agb.rst",
        "chains/fd.rst",
        "chains/fh.rst",
        "chains/tomo_fh.rst",
        "utilities/ground_cancellation.rst",
        "utilities/raster.rst",
        "utilities/plot.rst",
    ]
    for rst_file in rst_files:
        session.run(
            "python", "-m", "sphinx.ext.autosummary.generate", str(Path("doc/api").joinpath(rst_file)),
        )
    session.run(
        "python", "-m", "sphinx", "-b=html", "-a", "doc", "doc/_build",
    )



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import numpy as np
import os
import logging
from numpy import linalg as LA
from scipy.signal import medfilt2d
from osgeo import gdal
from osgeo.gdalconst import GA_ReadOnly

from biopal.statistics.utility_statistics import (
    main_correlation_estimation_SR,
    MPMBshuffle,
    main_correlation_estimation_SSF_SR,
)


def volume_decorrelation_lut(num_baselines, vertical_wavenumber, model_parameters):
    """
    Generation of a look-up table containing, for each value of height, extintion 
    and baseline, the possible values of volume decorrelation (the decorrelation caused
    by the different projection of the vertical component of the scatterer reflectivity 
    spectrum into the interferometric baselines images) [1][2]
    
    [1] A. Moreira, P. Prats-Iraola, M. Younis, G. Krieger, I. Hajnsek and K. P. Papathanassiou,"A tutorial on synthetic aperture radar," in IEEE Geoscience and Remote Sensing Magazine, vol. 1, no. 1, pp. 6-43, March 2013, doi: 10.1109/MGRS.2013.2248301.
    [2] K. P. Papathanassiou and S. R. Cloude, "Single-baseline polarimetric SAR interferometry," in IEEE Transactions on Geoscience and Remote Sensing, vol. 39, no. 11, pp. 2352-2363, Nov. 2001, doi: 10.1109/36.964971.
    
    """

    # LUT
    Nh = model_parameters.maximum_height  # 61
    Ns = model_parameters.number_of_extinction_value  # 51
    heights = np.arange(Nh)  # heights    [m]
    extinctions = np.concatenate(
        (np.tan(np.arange(Ns - 1) / (Ns - 1) * np.pi / 2.0) / 30.0, np.array([np.inf])), axis=0
    )  # extinctions    [m^(-1)]

    LUT = np.zeros((Nh, Ns, num_baselines), dtype=np.complex64)

    p = 2.0 * extinctions.reshape((Ns, 1)) * np.ones((1, num_baselines))
    q = np.ones((Ns, 1)) * (1j * vertical_wavenumber).reshape((1, num_baselines))
    q0 = q[0, :]
    p1 = p + q
    for hi in np.arange(1, Nh):  # do not cycle hi = 0
        LU = np.zeros((Ns, num_baselines), dtype=np.complex64)
        LU[1:-1, :] = (
            p[1:-1, :]
            / p1[1:-1, :]
            * (np.exp(p1[1:-1, :] * heights[hi]) - 1.0)
            / (np.exp(p[1:-1, :] * heights[hi]) - 1.0)
        )
        LU[0, :] = (np.exp(q0 * heights[hi]) - 1.0) / (q0 * heights[hi])  # for zero extinction
        LU[-1, :] = np.exp(q0 * heights[hi])  # for inifinite extinction
        LUT[hi, :, :] = LU

    LUT[0, :, :] = np.ones((Ns, num_baselines))  # zero height

    return LUT, extinctions


def estimate_height_core(PI, model_parameters, LUT, extinctions):
    """This function returns forest height, extinction, ground to volume ratio,
       temporal decorrelations.
    """
    num_baselines = PI.shape[2]
    Nh = model_parameters.maximum_height  # 61
    Ns = model_parameters.number_of_extinction_value  # 51
    heights = np.arange(Nh)  # heights    [m]
    heights[0] = 0.1  # to avoid dividing by zero in zero extinction

    if num_baselines == 1:
        # single baseline processing

        PI = np.squeeze(PI)
        gamma_v, mu, err_fro = get_Ground_and_Volume_components(PI, gamma_g=1, eps=1e-6)

        distances = np.abs(LUT[:, :, 0] - gamma_v) ** 2
        # distances to each (h, /sigma) pairs.

        subscribe = np.argmin(distances)

        # Pair of (h, /sigma) of the minimum distance.
        hgt = int(subscribe / Ns)
        ext = int(subscribe % Ns)

        # Assigning the results.
        output_height = heights[hgt]
        output_extinction = extinctions[ext]
        output_GVratio = mu

        output_gammaT1 = np.nan
        output_gammaT2 = np.nan
        output_gammaT3 = np.nan

        output = [
            output_height,
            output_extinction,
            output_GVratio,
            output_gammaT1,
            output_gammaT2,
            output_gammaT3,
        ]

    else:
        # multi baselines processing

        # The furthest point from the ground position is found.
        N_ch = PI.shape[0]
        num_baselines = PI.shape[2]

        I = np.eye(N_ch)  # Identity matrix
        gammav0 = np.zeros(num_baselines, dtype=np.complex64)
        for j in np.arange(num_baselines):
            P = PI[:, :, j] - I
            B = P @ np.conjugate(np.transpose(P))
            w, U = LA.eig(B)
            max_index = np.argmax(w, axis=0)
            w = U[:, max_index].reshape((1, N_ch))
            gammav0[j] = np.conjugate(w) @ P @ np.transpose(w)
            gammav0[j] = gammav0[j] + 1.0

        if np.sum(np.abs(gammav0) >= 1.0) >= 1:
            return [0, 0, np.inf, 1, 1, 1]

        # Solution Search
        # determine maximum mu
        a = np.abs(gammav0 - 1) ** 2
        b = 2.0 * np.real((gammav0 - 1) * np.conjugate(gammav0))
        c = np.abs(gammav0) ** 2 - 1.0
        mumax = np.min((-b + np.sqrt(b ** 2 - 4.0 * a * c)) / 2.0 / a)
        # Depending on baseline, the maximum mu differs.
        # The search should be performed only to the smallest value of the maximal mu's.

        Nmu = model_parameters.number_of_ground_volume_ratio_value  # 100 #number of mu's to be tested.
        Ngt = model_parameters.number_of_temporal_decorrelation_value  # 21  #number of gamma_T's to be tested. >1

        mini = np.zeros(Nmu)
        minima = np.zeros((Ngt, Ngt, Ngt))
        hgt = np.zeros((Ngt, Ngt, Ngt), dtype=np.int32)
        ext = np.zeros((Ngt, Ngt, Ngt), dtype=np.int32)

        for i in np.arange(Nmu):
            mu = i / Nmu * mumax
            gv = mu * (gammav0 - 1.0) + gammav0  # volume only points under given mu.
            distances = np.zeros((Nh, Ns))
            for j in np.arange(num_baselines):
                distances += np.abs(LUT[:, :, j] - gv[j]) ** 2
                # Sum of the distances to each (h, /sigma) pairs.

            mini[i] = np.min(distances)
        subscript = np.argmin(mini)
        mu = float(subscript) / Nmu * mumax
        gv = mu * (gammav0 - 1.0) + gammav0

        gammaT1 = np.arange(Ngt) / (Ngt - 1) * (1.0 - np.abs(gv[0])) + np.abs(gv[0])
        gammaT2 = np.arange(Ngt) / (Ngt - 1) * (1.0 - np.abs(gv[1])) + np.abs(gv[1])
        gammaT3 = np.arange(Ngt) / (Ngt - 1) * (1.0 - np.abs(gv[2])) + np.abs(gv[2])

        for gt1 in np.arange(Ngt):
            for gt2 in np.arange(Ngt):
                for gt3 in np.arange(Ngt):
                    distances = np.zeros((Nh, Ns))
                    gammav0gt = gv / [gammaT1[gt1], gammaT2[gt2], gammaT3[gt3]]
                    # gamma volumes after compensation of temporal decorreltaions.

                    for j in np.arange(num_baselines):
                        distances += np.abs(LUT[:, :, j] - gammav0gt[j]) ** 2
                    # Sum of the distances to each (h, /sigma) pairs.

                    minima[gt3, gt2, gt1] = np.min(distances)
                    subscribe = np.argmin(distances)

                    # Pair of (h, /sigma) of the minimum distance.
                    hgt[gt3, gt2, gt1] = subscribe / Ns
                    ext[gt3, gt2, gt1] = subscribe % Ns

        # Assigning the results.
        subscript = np.argmin(minima)
        pos = np.unravel_index(subscript, minima.shape)

        output_height = heights[hgt[pos]]
        output_extinction = extinctions[ext[pos]]
        output_GVratio = mu

        output_gammaT1 = gammaT1[pos[2]]
        output_gammaT2 = gammaT2[pos[1]]
        output_gammaT3 = gammaT3[pos[0]]

        output = [
            output_height,
            output_extinction,
            output_GVratio,
            output_gammaT1,
            output_gammaT2,
            output_gammaT3,
        ]

    return output


def estimate_height(
    data_stack,
    cov_est_window_size,
    pixel_spacing_slant_rg,
    pixel_spacing_az,
    incidence_angle_rad,
    carrier_frequency_hz,
    range_bandwidth_hz,
    vertical_wavenumber_stack,
    fh_proc_conf,
    R=None,
    look_angles=None,
    ground_slope=None,
):

    # data_stack is a dictionary of two nested dictionaries composed as:
    # data_stack[ acquisition_name ][ polarization ]

    num_acq = len(data_stack)
    acq_names = list(data_stack.keys())
    first_acq_dict = data_stack[acq_names[0]]
    pol_names = list(first_acq_dict.keys())
    num_pols = len(pol_names)
    Nrg, Naz = first_acq_dict[pol_names[0]].shape

    num_baselines = int((num_acq * (num_acq - 1)) / 2)

    # terrain = np.zeros( (Nrg, Naz) )
    if fh_proc_conf.spectral_shift_filtering:
        if R is None or look_angles is None or ground_slope is None:
            raise RuntimeError(
                'FH: when spectral shift filtering is enabled, "R", "look_angles" and "ground_slope" inputs should be specified.'
            )
        (MPMB_correlation, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_correlation_estimation_SSF_SR(
            data_stack,
            cov_est_window_size,
            pixel_spacing_slant_rg,
            pixel_spacing_az,
            incidence_angle_rad,
            R,
            look_angles,
            ground_slope,
            carrier_frequency_hz,
            range_bandwidth_hz,
        )

    else:
        if R is not None or look_angles is not None or ground_slope is not None:
            logging.warning(
                'FH: when spectral shift filtering is disabled, "R", "look_angles" and "ground_slope" inputs are not used.'
            )
        (MPMB_correlation, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_correlation_estimation_SR(
            data_stack,
            cov_est_window_size,
            pixel_spacing_slant_rg,
            pixel_spacing_az,
            incidence_angle_rad,
            carrier_frequency_hz,
            range_bandwidth_hz,
        )

    MBMP_correlation = MPMBshuffle(MPMB_correlation, rg_vec_subs, az_vec_subs, num_pols, num_acq)

    del MPMB_correlation

    Nrg_subs = rg_vec_subs.size
    Naz_subs = az_vec_subs.size
    heightmap = np.zeros((Nrg_subs, Naz_subs))
    extinctionmap = np.zeros((Nrg_subs, Naz_subs))
    ratiomap = np.zeros((Nrg_subs, Naz_subs))
    gammaT1map = np.zeros((Nrg_subs, Naz_subs))
    gammaT2map = np.zeros((Nrg_subs, Naz_subs))
    gammaT3map = np.zeros((Nrg_subs, Naz_subs))
    vertical_wavenumber = np.zeros((Nrg_subs, Naz_subs, num_baselines))

    Nrg_subs_string = str(Nrg_subs)
    for rg_sub_idx in np.arange(Nrg_subs):

        logging.info("   Heigth step " + str(rg_sub_idx + 1) + " of " + Nrg_subs_string)
        for az_sub_idx in np.arange(Naz_subs):

            # for az_sub_idx in np.arange(Naz_subs):
            current_vertical_wavenumber = np.zeros((num_acq, 1))
            for b_idx, stack_curr in enumerate(vertical_wavenumber_stack.values()):
                current_vertical_wavenumber[b_idx] = stack_curr[rg_vec_subs[rg_sub_idx], az_vec_subs[az_sub_idx]]

            current_vertical_wavenumber = current_vertical_wavenumber - current_vertical_wavenumber.T
            current_correlation = MBMP_correlation[:, :, rg_sub_idx, az_sub_idx]
            current_correlation[np.abs(current_correlation) > 1] = np.exp(
                1j * np.angle(current_correlation[np.abs(current_correlation) > 1])
            )

            PI = np.zeros((num_pols, num_pols, num_baselines), dtype=np.complex64)
            n = 0
            for i in np.arange(num_acq - 1):
                for j in np.arange(i + 1, num_acq):

                    PI[:, :, n] = current_correlation[
                        i * num_pols : num_pols + i * num_pols, j * num_pols : num_pols + j * num_pols,
                    ]
                    vertical_wavenumber[rg_sub_idx, az_sub_idx, n] = current_vertical_wavenumber[i, j]
                    n += 1

            if np.any(np.isnan(current_vertical_wavenumber)) + np.any(np.isnan(current_correlation)):
                heightmap[rg_sub_idx, az_sub_idx] = np.NaN
                extinctionmap[rg_sub_idx, az_sub_idx] = np.NaN
                ratiomap[rg_sub_idx, az_sub_idx] = np.NaN
                gammaT1map[rg_sub_idx, az_sub_idx] = np.NaN
                gammaT2map[rg_sub_idx, az_sub_idx] = np.NaN
                gammaT3map[rg_sub_idx, az_sub_idx] = np.NaN
                continue

            LUT, extinctions = volume_decorrelation_lut(
                PI.shape[2], vertical_wavenumber[rg_sub_idx, az_sub_idx, :], fh_proc_conf.model_parameters
            )

            output = estimate_height_core(PI, fh_proc_conf.model_parameters, LUT, extinctions)

            heightmap[rg_sub_idx, az_sub_idx] = output[0]
            extinctionmap[rg_sub_idx, az_sub_idx] = output[1]
            ratiomap[rg_sub_idx, az_sub_idx] = output[2]
            gammaT1map[rg_sub_idx, az_sub_idx] = output[3]
            gammaT2map[rg_sub_idx, az_sub_idx] = output[4]
            gammaT3map[rg_sub_idx, az_sub_idx] = output[5]

    logging.info("   performing median filtering of height products...")
    heightmap = medfilt2d(heightmap, kernel_size=fh_proc_conf.median_factor)
    extinctionmap = medfilt2d(extinctionmap, kernel_size=fh_proc_conf.median_factor)
    ratiomap = medfilt2d(ratiomap, kernel_size=fh_proc_conf.median_factor)
    gammaT1map = medfilt2d(gammaT1map, kernel_size=fh_proc_conf.median_factor)
    gammaT2map = medfilt2d(gammaT2map, kernel_size=fh_proc_conf.median_factor)
    gammaT3map = medfilt2d(gammaT3map, kernel_size=fh_proc_conf.median_factor)
    logging.info("    ...done.")

    return (
        heightmap,
        extinctionmap,
        ratiomap,
        gammaT1map,
        gammaT2map,
        gammaT3map,
        vertical_wavenumber,
        rg_vec_subs,
        az_vec_subs,
        subs_F_r,
        subs_F_a,
        MBMP_correlation,
    )


def heigths_masking_and_merging(data_equi7_fnames, mask_equi7_fnames, stacks_to_merge_dict):

    # for each stack:
    # 1) check if it is alone, of if there is a couple ASC+DES stacks
    # 2) if alone, gust apply the mask
    # 3) if it is not alone, merge it togheter with its companion data, also applying the mask

    # the asc_des_string is formatted as it follow:
    # XXX:SELF,YYY:STACK_ID
    # where:
    #   XXX can be 'ASC' or 'DES'
    #   YYY can be 'ASC' or 'DES' (if XXX='ASC', than YYY='DES' and vicecersa )
    #   'SELF' is a constant string: the asc_des_string is contained in a dictionary, 'SELF' is the dictionary key
    #   STACK_ID can be 'N.A' or can be the name of a stack_id

    # each input in data_equi7_fnames is a tiff with two layers
    # first later is data
    data_layer_index = 1
    # second layer is quality
    quality_layer_index = 2

    temp_path = next(iter(data_equi7_fnames.values()))[0]
    idx = temp_path.find("equi7")
    merging_folder = os.path.join(temp_path[: idx + len("equi7")], "merged")

    merged_data_fnames = []
    for (current_merging_id, unique_stack_ids_to_merge_list) in stacks_to_merge_dict.items():

        # prepare the empty dictionary for equi7 tile names
        equi7_tile_names = []

        # retrive the names of all the used "EQUI7 tiles"
        for curr_equi7_name in data_equi7_fnames[unique_stack_ids_to_merge_list[0]]:
            curr_equi7_tile_name = os.path.basename(os.path.dirname(curr_equi7_name))
            if not curr_equi7_tile_name in equi7_tile_names:
                equi7_tile_names.append(curr_equi7_tile_name)

        first_stack_data_fname = data_equi7_fnames[unique_stack_ids_to_merge_list[0]][0]
        data_tiff_name_in = os.path.basename(first_stack_data_fname)

        data_tiff_name_out = data_tiff_name_in.replace(unique_stack_ids_to_merge_list[0], current_merging_id)

        p1 = os.path.dirname(os.path.dirname(first_stack_data_fname))
        equi7_subgridname = os.path.basename(p1)

        for equi7_tile_name in equi7_tile_names:

            out_data_fname = os.path.join(
                merging_folder, current_merging_id, equi7_subgridname, equi7_tile_name, data_tiff_name_out,
            )

            os.makedirs(os.path.dirname(out_data_fname))

            for idx, unique_stack_id in enumerate(unique_stack_ids_to_merge_list):
                # input and output paths

                curr_data_fname = [name for name in data_equi7_fnames[unique_stack_id] if equi7_tile_name in name][0]
                curr_mask_fname = [name for name in mask_equi7_fnames[unique_stack_id] if equi7_tile_name in name][0]

                # Load DATA
                data_driver = gdal.Open(curr_data_fname, GA_ReadOnly)
                data_curr = data_driver.GetRasterBand(data_layer_index).ReadAsArray()
                projection = data_driver.GetProjection()
                geotransform = data_driver.GetGeoTransform()
                # Load quality layer
                quality_data_curr = data_driver.GetRasterBand(quality_layer_index).ReadAsArray()
                data_driver = None

                # Load MASK
                mask_driver = gdal.Open(curr_mask_fname, GA_ReadOnly)
                mask_curr = (mask_driver.GetRasterBand(1).ReadAsArray()).astype("bool")
                mask_driver = None

                if idx == 0:
                    FH_out_data = data_curr
                    quality_out_data = quality_data_curr

                    FH_out_data[np.logical_not(mask_curr)] = np.nan
                    quality_out_data[np.logical_not(mask_curr)] = np.nan

                    mask_prev = mask_curr

                else:

                    data_prev = FH_out_data
                    quality_data_prev = quality_out_data
                    del FH_out_data, quality_out_data

                    # 1) mean of the current and previous:
                    data_out_curr = (data_curr + data_prev) / 2
                    quality_data_out_curr = (quality_data_curr + quality_data_prev) / 2

                    # 2) where current only is valid use current values and viceversa
                    curr_only_valid_idxes = np.logical_and(mask_curr, np.logical_not(mask_prev))
                    prev_only_valid_idxes = np.logical_and(mask_prev, np.logical_not(mask_curr))

                    data_out_curr[curr_only_valid_idxes] = data_curr[curr_only_valid_idxes]
                    data_out_curr[prev_only_valid_idxes] = data_prev[prev_only_valid_idxes]
                    quality_data_out_curr[curr_only_valid_idxes] = quality_data_curr[curr_only_valid_idxes]
                    quality_data_out_curr[prev_only_valid_idxes] = quality_data_prev[prev_only_valid_idxes]

                    # 3) where none are valid, write NAN
                    invalid_idxes = np.logical_and(np.logical_not(mask_prev), np.logical_not(mask_curr))

                    data_out_curr[invalid_idxes] = np.nan
                    quality_data_out_curr[invalid_idxes] = np.nan

                    # 4) keep only common parts:
                    not_common_parts_indexes = np.logical_or(np.isnan(data_curr), np.isnan(data_prev))

                    data_out_curr[not_common_parts_indexes] = np.nan
                    quality_data_out_curr[not_common_parts_indexes] = np.nan

                    FH_out_data = data_out_curr
                    quality_out_data = quality_data_out_curr

                    mask_prev = mask_curr

                    del data_out_curr, quality_data_out_curr, data_prev, quality_data_prev

            # save to file:
            Nx, Ny = FH_out_data.shape

            driver = gdal.GetDriverByName("GTiff")
            outdata = driver.Create(out_data_fname, Ny, Nx, 2, gdal.GDT_Float32)
            outdata.SetGeoTransform(geotransform)  ##sets same geotransform as input
            outdata.SetProjection(projection)  ##sets same projection as input
            outdata.GetRasterBand(data_layer_index).WriteArray(FH_out_data)
            outdata.GetRasterBand(quality_layer_index).WriteArray(quality_out_data)
            outdata.FlushCache()  ##saves to disk!!
            outdata = None

        merged_data_fnames.append(out_data_fname)
        logging.info("    ...done.")

    return merged_data_fnames, merging_folder


def get_volume_SB_Pi(Pi, gamma_g, eps=1e-6):
    """
    Extracts the whitened volume component corresponding to the minimum
    Entropy ground solution (rank-2 ground) and the volume coherence
    (gamma_volume) from the given whitened PolInSAr matrix Pi and the
    ground coherence gamma_g.

    In order to avoid numerical errors the minimum eigenvalue of the whitened
    ground is set to eps instead of 0.

    Parameters
    ----------
    Pi : ndarray (3,3)
        Whitened PolInSAR matrix
    gamma_g : complex scalar
        Ground coherence (gamma_ground)
    eps : scalar, optional
        Power minimum ground eigenvalue (whitened). Set it to a small value
        to avoid numerical problems when its equal to 0.

    Returns
    -------
    Tvw : ndarray (3,3)
        Whitened volume component corresponding to a rank-2 ground component
    gamma_v : complex scalar
        Selected volume coherence (gamma_volume)
    """
    # Get center of mass coherence region
    gamma_c = np.trace(Pi) / 3
    # Get whitened volume with gamma_c
    Tvh = (Pi - gamma_g * np.eye(3)) / (gamma_c - gamma_g)
    Tvh = 0.5 * (Tvh + np.conj(Tvh.T))
    # Scale it to max eigenvalue (1.0 - eps)
    l1 = np.max(np.linalg.eigvalsh(Tvh))
    # Get obtained gamma_volume
    gamma_v = gamma_g + (gamma_c - gamma_g) * l1 / (1.0 - eps)
    return Tvh * (1.0 - eps) / l1, gamma_v


def get_Ground_and_Volume_components(Pi, gamma_g=1, eps=1e-6):
    """
    Computes the Ground and Volume components
    corresponding to the minimum Entropy ground solution (rank-2 ground)
    and the relative error.

    Parameters
    ----------
    Pi : ndarray (3,3)
        Whitened PolInSAR matrix
    gamma_g : scalar, optional
        Ground coherence
    eps : scalar, optional
        Power minimum ground eigenvalue (whitened). Set it to a small value
        to avoid numerical problems when its equal to 0.

    Returns
    -------
    gamma_v : complex scalar
        Selected volume coherence (gamma_volume)
    mu : scalar
        minimum ground to volume ratio
    err_fro : scalar
        Frobenius relative squared error performed in the 2 layer modelling
        with respect to original data
    """
    # Get whitened volume component & gamma_volume
    Tvw, gamma_v = get_volume_SB_Pi(Pi, gamma_g, eps)

    # Get whitened ground component
    Tgw = np.eye(3) - Tvw

    # Compute Modelled PolInSAR coherence (within the line)
    Pi_m = gamma_g * Tgw + gamma_v * Tvw

    # Compute Frobenius relative modelling squared error
    err_fro = (np.linalg.norm(Pi - Pi_m) / np.linalg.norm(Pi)) ** 2

    # Minimum Ground-to-volume ratio
    mu = np.min(np.diag(Tgw / Tvw))

    return gamma_v, mu, err_fro



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import numpy as np
from scipy.interpolate import interp1d
from biopal.io.xml_io import raster_info


def apply_calibration_screens(beta0_calibrated, beta0_raster_info, cal_screens, screens_raster_info, master_id):

    for idx, pf_name in enumerate(beta0_calibrated.keys()):

        if idx == 0:
            rg_axis_index = 0
            az_axis_index = 1

            # input original axis
            max_val_in_rg = screens_raster_info.num_samples * screens_raster_info.pixel_spacing_slant_rg
            rg_ax_in = np.arange(0, max_val_in_rg, raster_info.pixel_spacing_slant_rg)

            max_val_in_az = screens_raster_info.num_lines * screens_raster_info.pixel_spacing_az
            az_ax_in = np.arange(0, max_val_in_az, screens_raster_info.pixel_spacing_az)

            # output interpolated axis
            max_val_out_rg = beta0_raster_info.num_samples * beta0_raster_info.pixel_spacing_slant_rg
            rg_ax_out = np.arange(0, max_val_out_rg, beta0_raster_info.pixel_spacing_slant_rg)

            max_val_out_az = beta0_raster_info.num_lines * beta0_raster_info.pixel_spacing_az
            az_ax_out = np.arange(0, max_val_out_az, beta0_raster_info.pixel_spacing_az)

        # interpolate current screen over the data definiton axes:
        cal_scree_interp = cal_screens[pf_name]

        # interpolation of data along axis_index
        interp_fun = interp1d(rg_ax_in, cal_scree_interp, rg_axis_index, bounds_error=0)
        cal_scree_interp = interp_fun(rg_ax_out)
        interp_fun = interp1d(az_ax_in, cal_scree_interp, az_axis_index, bounds_error=0)
        cal_scree_interp = interp_fun(az_ax_out)

        for pol_id in beta0_calibrated[pf_name]:
            beta0_calibrated[pf_name][pol_id] = np.multiply(
                beta0_calibrated[pf_name][pol_id], np.exp(-1j * cal_scree_interp)
            )

    return beta0_calibrated
